/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
 */


//
//  FCompoundFile.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFComparable.h"
#include "FSectorCollection.h"
#include "FDirectoryEntry.h"
#include "FHeader.h"

#include "RBT/BTSearchTree.h"
#include "RBT/BTNode.h"
#include "FItem.h"
#include "FStreamView.h"
#include "FException.h"

//#define FLAT_WRITE // No optimization on the number of write operations

class CFStream;
class CFStorage;

class CFItemComparer : public ITComparer<CFItem>
{
public:
	int Compare(CFItem* x, CFItem* y);
};

// Binary File Format Version. Sector size  is 512 byte for version 3, 4096 for version 4
typedef enum _tagCFVersion
{
	// Compound file version 3 - The default and most common version available. Sector size 512 bytes, 2GB max file size.
	eVer_3 = 3,
	// Compound file version 4 - Sector size is 4096 bytes. Using this version could bring some compatibility problem with existing applications.
	eVer_4 = 4
}ECFVersion;

// Update mode of the compound file. Default is ReadOnly.
typedef enum _tagCFUpdateMode
{
    /// ReadOnly update mode prevents overwriting
    /// of the opened file. 
    /// Data changes are allowed but they have to be 
    /// persisted on a different file when required.
    eReadOnly,
    /// Update mode allows subsequent data changing operations
    /// to be persisted directly on the opened file or stream
    /// using the <see cref="M:OpenMcdf.CompoundFile.Commit">Commit</see>
    /// method when required. Warning: this option may cause existing data loss if misused.
    eUpdate
}ECFUpdateMode;

class LIBCDF_EXPORT CFCompoundFile : public RefObject
{
public:
	CFCompoundFile(void);
	virtual ~CFCompoundFile(void);

	/// <summary>
	/// Number of DIFAT entries in the _header
	/// </summary>
	static int HEADER_DIFAT_ENTRIES_COUNT;// = 109;

	/// <summary>
	/// Number of FAT entries in a DIFAT Sector
	/// </summary>
	static int DIFAT_SECTOR_FAT_ENTRIES_COUNT;// = 127;

	/// <summary>
	/// Sectors ID entries in a FAT Sector
	/// </summary>
	static int FAT_SECTOR_ENTRIES_COUNT;// = 128;

	/// <summary>
	/// Sector ID Size (int)
	/// </summary>
	static int SIZE_OF_SID;// = 4;

	/// <summary>
	/// Initial capacity of the flushing queue used
	/// to optimize commit writing operations
	/// </summary>
	static int FLUSHING_QUEUE_SIZE;// = 6000;

	/// <summary>
	/// Maximum size of the flushing buffer used
	/// to optimize commit writing operations
	/// </summary>
	static int FLUSHING_BUFFER_MAX_SIZE;// = 1024 * 1024 * 16;

    /// <summary>
    /// Returns the size of standard sectors switching on CFS version (3 or 4)
    /// </summary>
    /// <returns>Standard sector size</returns>
    int GetSectorSize()
    {
        return 2 << (_header->SectorShift() - 1);
    }
    
    const std::wstring& GetFileName() const { return _fileName; }

private:
    /// <summary>
    /// Flag for sector recycling.
    /// </summary>
	bool _sectorRecycle;// = false;

    /// <summary>
    /// Flag for unallocated sector zeroing out.
    /// </summary>
    bool _eraseFreeSectors;// = false;


    CFSectorCollection* _sectors;// = new SectorCollection();

    /// <summary>
    /// CompoundFile _header
    /// </summary>
    CFHeader* _header;

    /// <summary>
    /// Compound underlying stream-> Null when new CF has been created.
    /// </summary>
    CFileStream* _sourceStream;

    ECFUpdateMode _updateMode;// = ECFUpdateMode.ReadOnly;
    std::wstring _fileName;// = String.Empty;
    bool _validationExceptionEnabled;// = true;
    
    bool _dirty;    //  dirty flag

	 
#ifndef FLAT_WRITE
	CFByteArray _buffer;

    SectorQueue _flushingQueue;// = new Queue<Sector>(FLUSHING_QUEUE_SIZE);
#endif

	void Init();

public:

    void OnSizeLimitReached();

    /// <summary>
    /// Create a new, blank, compound file.
    /// </summary>
    /// <param name="cfsVersion">Use a specific Compound File Version to set 512 or 4096 bytes sectors</param>
    /// <param name="_sectorRecycle">If true, recycle unused sectors</param>
    /// <param name="_eraseFreeSectors">If true, unallocated sectors will be overwritten with zeros</param>
    /// <example>
    /// <code>
    /// 
    ///     byte[] b = new byte[10000];
    ///     for (int i = 0; i &lt; 10000; i++)
    ///     {
    ///         b[i % 120] = (byte)i;
    ///     }
    ///
    ///     CompoundFile cf = new CompoundFile(CFSVersion.Ver_4, true, true);
    ///     CFStream myStream = cf.RootStorage.AddStream("MyStream");
    ///
    ///     Assert.IsNotNull(myStream);
    ///     myStream.SetData(b);
    ///     cf.Save("MyCompoundFile.cfs");
    ///     cf.Close();
    ///     
    /// </code>
    /// </example>
    /// <remarks>
    /// Sector recycling reduces data writing performances but avoids space wasting in scenarios with frequently
    /// data manipulation of the same streams. The new compound file is open in Update mode.
    /// </remarks>
    CFCompoundFile(ECFVersion cfsVersion, bool sectorRecycle, bool eraseFreeSectors);

    /// <summary>
    /// Load an existing compound file.
    /// </summary>
    /// <param name="fileName">Compound file to read from</param>
    /// <example>
    /// <code>
    /// //A xls file should have a Workbook stream
    /// String filename = "report.xls";
    ///
    /// CompoundFile cf = new CompoundFile(filename);
    /// CFStream foundStream = cf.RootStorage.GetStream("Workbook");
    ///
    /// byte[] temp = foundStream.GetData();
    ///
    /// Assert.IsNotNull(temp);
    ///
    /// cf.Close();
    /// </code>
    /// </example>
    /// <remarks>
    /// File will be open in read-only mode: it has to be saved
    /// with a different filename. A wrapping implementation has to be provided 
    /// in order to remove/substitute an existing file. Version will be
    /// automatically recognized from the file. Sector recycle is turned off
    /// to achieve the best reading/writing performance in most common scenarios.
    /// </remarks>
    CFCompoundFile(const std::wstring& fileName);

    /// <summary>
    /// Load an existing compound file.
    /// </summary>
    /// <param name="fileName">Compound file to read from</param>
    /// <param name="_sectorRecycle">If true, recycle unused sectors</param>
    /// <param name="updateMode">Select the update mode of the underlying data file</param>
    /// <param name="_eraseFreeSectors">If true, overwrite with zeros unallocated sectors</param>
    /// <example>
    /// <code>
    /// String srcFilename = "data_YOU_CAN_CHANGE.xls";
    /// 
    /// CompoundFile cf = new CompoundFile(srcFilename, ECFUpdateMode.Update, true, true);
    ///
    /// Random r = new Random();
    ///
    /// byte[] buffer = GetBuffer(r.Next(3, 4095), 0x0A);
    ///
    /// cf.RootStorage.AddStream("MyStream").SetData(buffer);
    /// 
    /// //This will persist data to the underlying media.
    /// cf.Commit();
    /// cf.Close();
    ///
    /// </code>
    /// </example>
    CFCompoundFile(const std::wstring& fileName, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors);

    /// <summary>
    /// Load an existing compound file.
    /// </summary>
    /// <param name="fileName">Compound file to read from</param>
    /// <param name="_sectorRecycle">If true, recycle unused sectors</param>
    /// <param name="updateMode">Select the update mode of the underlying data file</param>
    /// <param name="_eraseFreeSectors">If true, overwrite with zeros unallocated sectors</param>
    /// <param name="noValidationException">If true, no <see cref="T:OpenMcdf.std::exception">std::exception</see>
    ///  will be thrown even if corrupted file is loaded. Please note that this option is could pose a potential security threat</param>
    /// <example>
    /// <code>
    /// String srcFilename = "data_YOU_CAN_CHANGE.xls";
    /// 
    /// CompoundFile cf = new CompoundFile(srcFilename, ECFUpdateMode.Update, true, true, true);
    ///
    /// 
    /// Random r = new Random();
    ///
    /// byte[] buffer = GetBuffer(r.Next(3, 4095), 0x0A);
    ///
    /// cf.RootStorage.AddStream("MyStream").SetData(buffer);
    /// 
    /// //This will persist data to the underlying media.
    /// cf.Commit();
    /// cf.Close();
    ///
    /// </code>
    /// </example>
    CFCompoundFile(const std::wstring& fileName, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors, bool noValidationException);


    PROPERTY_GET(bool, ValidationExceptionEnabled, _validationExceptionEnabled);

    /// <summary>
    /// Load an existing compound file.
    /// </summary>
    /// <param name="stream">A stream containing a compound file to read</param>
    /// <param name="_sectorRecycle">If true, recycle unused sectors</param>
    /// <param name="updateMode">Select the update mode of the underlying data file</param>
    /// <param name="_eraseFreeSectors">If true, overwrite with zeros unallocated sectors</param>
    /// <example>
    /// <code>
    /// 
    /// String filename = "reportREAD.xls";
    ///   
    /// FileStream fs = new FileStream(filename, FileMode.Open);
    /// CompoundFile cf = new CompoundFile(fs, ECFUpdateMode.ReadOnly, false, false);
    /// CFStream foundStream = cf.RootStorage.GetStream("Workbook");
    ///
    /// byte[] temp = foundStream.GetData();
    ///
    /// Assert.IsNotNull(temp);
    ///
    /// cf.Close();
    ///
    /// </code>
    /// </example>
    /// <exception cref="T:OpenMcdf.std::exception">Raised when trying to open a non-seekable stream</exception>
    /// <exception cref="T:OpenMcdf.std::exception">Raised stream is NULL</exception>
    CFCompoundFile(CFileStream* stream, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors);

    /// <summary>
    /// Load an existing compound file.
    /// </summary>
    /// <param name="stream">A stream containing a compound file to read</param>
    /// <param name="_sectorRecycle">If true, recycle unused sectors</param>
    /// <param name="updateMode">Select the update mode of the underlying data file</param>
    /// <param name="_eraseFreeSectors">If true, overwrite with zeros unallocated sectors</param>
    /// <param name="noValidationException">If true, openMcdf will try to ignore invalid references or format in order to load a possibly corrupted file anyway </param>
    /// <remarks>The 'noValidationEcxception' parameter could possibly lead to security issues so it's recommanded to use it only on trusted sources</remarks>
    /// <example>
    /// <code>
    /// 
    /// String filename = "reportREAD.xls";
    ///   
    /// FileStream fs = new FileStream(filename, FileMode.Open);
    /// CompoundFile cf = new CompoundFile(fs, ECFUpdateMode.ReadOnly, false, false, false, true);
    /// CFStream foundStream = cf.RootStorage.GetStream("Workbook");
    ///
    /// byte[] temp = foundStream.GetData(); //if 'reportRead.xls' is corrupted, openMcdf will try to lad it anyway [noValidationException set true]
    ///
    /// Assert.IsNotNull(temp);
    ///
    /// cf.Close();
    ///
    /// </code>
    /// </example>
    /// <exception cref="T:OpenMcdf.std::exception">Raised when trying to open a non-seekable stream</exception>
    /// <exception cref="T:OpenMcdf.std::exception">Raised stream is NULL</exception>
    CFCompoundFile(CFileStream* stream, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors, bool noValidationException);

    /// <summary>
    /// Load an existing compound file from a stream->
    /// </summary>
    /// <param name="stream">Streamed compound file</param>
    /// <example>
    /// <code>
    /// 
    /// String filename = "reportREAD.xls";
    ///   
    /// FileStream fs = new FileStream(filename, FileMode.Open);
    /// CompoundFile cf = new CompoundFile(fs);
    /// CFStream foundStream = cf.RootStorage.GetStream("Workbook");
    ///
    /// byte[] temp = foundStream.GetData();
    ///
    /// Assert.IsNotNull(temp);
    ///
    /// cf.Close();
    ///
    /// </code>
    /// </example>
    /// <exception cref="T:OpenMcdf.std::exception">Raised when trying to open a non-seekable stream</exception>
    /// <exception cref="T:OpenMcdf.std::exception">Raised stream is NULL</exception>
    CFCompoundFile(CFileStream* stream);


    /// <summary>
    /// Commit data changes since the previously commit operation
    /// to the underlying supporting stream or file on the disk.
    /// </summary>
    /// <remarks>
    /// This method can be used
    /// only if the supporting stream has been opened in 
    /// <see cref="T:OpenMcdf.ECFUpdateMode">Update mode</see>.
    /// </remarks>
    void Commit();

    /// <summary>
    /// Commit data changes since the previously commit operation
    /// to the underlying supporting stream or file on the disk.
    /// </summary>
    /// <param name="releaseMemory">If true, release loaded sectors to limit memory usage but reduces following read operations performance</param>
    /// <remarks>
    /// This method can be used only if 
    /// the supporting stream has been opened in 
    /// <see cref="T:OpenMcdf.ECFUpdateMode">Update mode</see>.
    /// </remarks>
    void Commit(bool releaseMemory);

    //  IFObject
public:
    virtual void Delete();
    virtual bool Equals(const IFObject* obj) const;

private:
    /// <summary>
    /// Load compound file from an existing stream->
    /// </summary>
    /// <param name="stream">Stream to load compound file from</param>
	void Load(CFileStream* stream);

    void LoadFile(const std::wstring& fileName);

    void LoadStream(CFileStream* stream);

public:
    /// <summary>
    /// Return true if this compound file has been 
    /// loaded from an existing file or stream
    /// </summary>
    bool HasSourceStream()
    {
        return _sourceStream != NULL;
    }

	CFileStream& SourceStream() {return *_sourceStream;}

protected:
    /// <summary>
    /// Allocate space, setup sectors id and refresh _header
    /// for the new or updated mini sector chain.
    /// </summary>
    /// <param name="sectorChain">The new MINI sector chain</param>
    void SetMiniSectorChain(SectorChain* sectorChain);

    /// <summary>
    /// Allocate space, setup sectors id and refresh _header
    /// for the new or updated sector chain.
    /// </summary>
    /// <param name="sectorChain">The new or updated generic sector chain</param>
    void SetNormalSectorChain(SectorChain* sectorChain);

	void FreeChain(SectorChain* sectorChain, bool zeroSector);

    void FreeMiniChain(SectorChain* sectorChain, bool zeroSector);

protected:
    bool _transactionLockAdded;// = false;
    int _lockSectorId;// = -1;
    bool _transactionLockAllocated;// = false;

public:
    /// <summary>
    /// Check for transaction lock sector addition and mark it in the FAT.
    /// </summary>
    void CheckForLockSector();

    /// <summary>
    /// Allocate space, setup sectors id and refresh _header
    /// for the new or updated FAT sector chain.
    /// </summary>
    /// <param name="sectorChain">The new or updated generic sector chain</param>
    void SetFATSectorChain(SectorChain* sectorChain);

    /// <summary>
    /// Setup the DIFAT sector chain
    /// </summary>
    /// <param name="FATsectorChain">A FAT sector chain</param>
    void SetDIFATSectorChain(SectorChain* sectorChain);


    /// <summary>
    /// Get the DIFAT Sector chain
    /// </summary>
    /// <returns>A list of DIFAT sectors</returns>
    SectorChain* GetDifatSectorChain();

    /// <summary>
    /// Get the FAT sector chain
    /// </summary>
    /// <returns>List of FAT sectors</returns>
    SectorChain* GetFatSectorChain();

    /// <summary>
    /// Get a standard sector chain
    /// </summary>
    /// <param name="secID">First SecID of the required chain</param>
    /// <returns>A list of sectors</returns>
    SectorChain* GetNormalSectorChain(int secID);

    /// <summary>
    /// Get a mini sector chain
    /// </summary>
    /// <param name="secID">First SecID of the required chain</param>
    /// <returns>A list of mini sectors (64 bytes)</returns>
    SectorChain* GetMiniSectorChain(int secID);

    /// <summary>
    /// Get a sector chain from a compound file given the first sector ID
    /// and the required sector type.
    /// </summary>
    /// <param name="secID">First chain sector's id </param>
    /// <param name="chainType">Type of Sectors in the required chain (mini sectors, normal sectors or FAT)</param>
    /// <returns>A list of Sectors as the result of their concatenation</returns>
	SectorChain* GetSectorChain(int secID, CFSector::ESectorType chainType);

    /// <summary>
    /// The entry point object that represents the 
    /// root of the structures tree to get or set storage or
    /// stream data.
    /// </summary>
    /// <example>
    /// <code>
    /// 
    ///    //Create a compound file
    ///    string FILENAME = "MyFileName.cfs";
    ///    CompoundFile ncf = new CompoundFile();
    ///
    ///    CFStorage l1 = ncf.RootStorage.AddStorage("Storage Level 1");
    ///
    ///    l1.AddStream("l1ns1");
    ///    l1.AddStream("l1ns2");
    ///    l1.AddStream("l1ns3");
    ///    CFStorage l2 = l1.AddStorage("Storage Level 2");
    ///    l2.AddStream("l2ns1");
    ///    l2.AddStream("l2ns2");
    ///
    ///    ncf.Save(FILENAME);
    ///    ncf.Close();
    /// </code>
    /// </example>
    CFStorage* RootStorage();

    ECFVersion Version()
    {
        return (ECFVersion)_header->MajorVersion();
    }

    void InsertNewDirectoryEntry(IDirectoryEntry* de);

    /// <summary>
    /// Reset a directory entry setting it to StgInvalid in the Directory.
    /// </summary>
    /// <param name="sid">Sid of the directory to invalidate</param>
    void ResetDirectoryEntry(int sid);

	bool RefreshSIDs(BinaryTrees::BinaryTreeNode<CFItem>& node);

    BTSearchTree<CFItem>* GetChildrenTree(int sid);

private:
	void DoLoadChildren(BTSearchTree<CFItem>& bst, IDirectoryEntry* de);

	// Doubling methods allows iterative behavior while avoiding
    // to insert duplicate items
	void LoadSiblings(BTSearchTree<CFItem>& bst, IDirectoryEntry* de);

private:
	void DoLoadSiblings(BTSearchTree<CFItem>& bst, IDirectoryEntry* de);

    bool ValidateSibling(int sid);

    /// <summary>
    /// Load directory entries from compound file. Header and FAT MUST be already loaded.
    /// </summary>
    void LoadDirectories();

public:
    void RefreshIterative(BinaryTrees::BinaryTreeNode<CFItem>* node);

public:
    bool IsDirty() const;
    void SetModifiedFlag(bool bModified = true);
    
    /// <summary>
    /// Saves the in-memory image of Compound File to a file.
    /// </summary>
    /// <param name="fileName">File name to write the compound file to</param>
    /// <exception cref="T:OpenMcdf.std::exception">Raised if destination file is not seekable</exception>
    
    void Save(const std::wstring& fileName, bool bUpdateFileName);
    
    
private:
    /// <summary>
    ///  Commit directory entries change on the Current Source stream
    /// </summary>
    void CommitDirectory();

    /// Save to the file (_fileName)
    void Save();

    /// <summary>
    /// Saves the in-memory image of Compound File to a stream->
    /// </summary>        
    /// <remarks>
    /// Destination Stream must be seekable.
    /// </remarks>
    /// <param name="stream">The stream to save compound File to</param>
    /// <exception cref="T:OpenMcdf.std::exception">Raised if destination stream is not seekable</exception>
    /// <exception cref="T:OpenMcdf.std::exception">Raised if Compound File Storage has been already disposed</exception>
    /// <example>
    /// <code>
    ///    MemoryStream ms = new MemoryStream(size);
    ///
    ///    CompoundFile cf = new CompoundFile();
    ///    CFStorage st = cf.RootStorage.AddStorage("MyStorage");
    ///    CFStream sm = st.AddStream("MyStream");
    ///
    ///    byte[] b = new byte[]{0x00,0x01,0x02,0x03};
    ///
    ///    sm.SetData(b);
    ///    cf.Save(ms);
    ///    cf.Close();
    /// </code>
    /// </example>
    void Save(CFileStream* stream);

public:
    /// <summary>
    /// Scan FAT o miniFAT for free sectors to reuse.
    /// </summary>
    /// <param name="sType">Type of sector to look for</param>
    /// <returns>A stack of available sectors or minisectors already allocated</returns>
    SectorQueue* FindFreeSectors(CFSector::ESectorType sType);

    void SetData(CFItem* cfItem, CFByteArray& buffer);

    /// <summary>
    /// INTERNAL DEVELOPMENT. DO NOT CALL.
    /// <param name="directoryEntry"></param>
    /// <param name="buffer"></param>
    void AppendData(CFItem* cfItem, CFByteArray& buffer);

private:
	void SetStreamData(CFItem* cfItem, CFByteArray& buffer);

public:

    unsigned char* GetData(CFStream* cFStream, long offset, int& count, CFByteArray& buffer);

    unsigned char* GetData(CFStream* cFStream, CFByteArray& buffer);

private:
    /// <summary>
    /// Check file size limit ( 2GB for version 3 )
    /// </summary>
    void CheckFileLength()
    {
        throw NotImplementedException();
    }

	static int Ceiling(double d)
    {
        return (int)ceil(d);
    }

    static int LowSaturation(int i)
    {
        return i > 0 ? i : 0;
    }

public:
    void RemoveDirectoryEntry(int sid);

 public:
   /// <summary>
    /// Close the Compound File object <see cref="T:OpenMcdf.CompoundFile">CompoundFile</see> and
    /// free all associated resources (e.g. open file handle and allocated memory).
    /// <remarks>
    /// When the <see cref="T:OpenMcdf.CompoundFile.Close()">Close</see> method is called,
    /// all the associated stream and storage objects are invalidated:
    /// any operation invoked on them will produce a <see cref="T:OpenMcdf.std::exception">std::exception</see>.
    /// </remarks>
    /// </summary>
    /// <example>
    /// <code>
    ///    const String FILENAME = "CompoundFile.cfs";
    ///    CompoundFile cf = new CompoundFile(FILENAME);
    ///
    ///    CFStorage st = cf.RootStorage.GetStorage("MyStorage");
    ///    cf.Close();
    ///
    ///    try
    ///    {
    ///        byte[] temp = st.GetStream("MyStream").GetData();
    ///        
    ///        // The following line will fail because back-end object has been closed
    ///        Assert.Fail("Stream without media");
    ///    }
    ///    catch (Exception ex)
    ///    {
    ///        Assert.IsTrue(ex is std::exception);
    ///    }
    /// </code>
    /// </example>
    void Close()
    {
        Dispose();
    }

    void Close(bool closeStream)
    {
		_closeStream = closeStream;
        Dispose();
    }

protected:

    void Dispose()
    {
        Dispose(true);
        //GC.SuppressFinalize(this);
    }

private:
	CFStorage* _rootStorage;

    bool _closeStream;// = true;
	bool _freeStream;// = false;


    bool _disposed;//false

    IFObject* _lockObject;// = new IFObject(); no use.
    
	CFDirectoryEntries* _directoryEntries;
	//   = new List<IDirectoryEntry>();


protected:
    /// <summary>
    /// When called from user code, release all resources, otherwise, in the case runtime called it,
    /// only unmanagd resources are released.
    /// </summary>
    /// <param name="disposing">If true, method has been called from User code, if false it's been called from .net runtime</param>
    virtual void Dispose(bool disposing);

public:
    bool IsClosed() const
    {
        return _disposed;
    }

    //List<IDirectoryEntry> DirectoryEntries
    //{
    //    get { return _directoryEntries; }
    //}

    IDirectoryEntry* RootEntry()
    {
        return (*_directoryEntries)[0];
    }

private:
	CFDirectoryEntries* FindDirectoryEntries(std::wstring entryName);

public:
    /// <summary>
    /// Get a list of all entries with a given name contained in the document.
    /// </summary>
    /// <param name="entryName">Name of entries to retrive</param>
    /// <returns>A list of name-matching entries</returns>
    /// <remarks>This function is aimed to speed up entity lookup in 
    /// flat-structure files (only one or little more known entries)
    /// without the performance penalty related to entities hierarchy constraints.
    /// There is no implied hierarchy in the returned list.
    /// </remarks>
    CFItemVector* GetAllNamedEntries(std::wstring entryName);
	static void ReleaseAllEntries(CFItemVector* pEntries);

	/// <summary>
    /// Compress free space by removing unallocated sectors from compound file
    /// effectively reducing stream or file size.
    /// </summary>
    /// <remarks>
    /// Current implementation supports compression only for ver. 3 compound files.
    /// </remarks>
    /// <example>
    /// <code>
    /// 
    ///  //This code has been extracted from unit test
    ///  
    ///    String FILENAME = "MultipleStorage3.cfs";
    ///
    ///    FileInfo srcFile = new FileInfo(FILENAME);
    ///
    ///    File.Copy(FILENAME, "MultipleStorage_Deleted_Compress.cfs", true);
    ///
    ///    CompoundFile cf = new CompoundFile("MultipleStorage_Deleted_Compress.cfs", ECFUpdateMode.Update, true, true);
    ///
    ///    CFStorage st = cf.RootStorage.GetStorage("MyStorage");
    ///    st = st.GetStorage("AnotherStorage");
    ///    
    ///    Assert.IsNotNull(st);
    ///    st.Delete("Another2Stream"); //17Kb
    ///    cf.Commit();
    ///    cf.Close();
    ///
    ///    CompoundFile.ShrinkCompoundFile("MultipleStorage_Deleted_Compress.cfs");
    ///
    ///    FileInfo dstFile = new FileInfo("MultipleStorage_Deleted_Compress.cfs");
    ///
    ///    Assert.IsTrue(srcFile.Length > dstFile.Length);
    ///
    /// </code>
    /// </example>
    static void ShrinkCompoundFile(CFileStream* s);

    /// <summary>
    /// Remove unallocated sectors from compound file in order to reduce its size.
    /// </summary>
    /// <remarks>
    /// Current implementation supports compression only for ver. 3 compound files.
    /// </remarks>
    /// <example>
    /// <code>
    /// 
    ///  //This code has been extracted from unit test
    ///  
    ///    String FILENAME = "MultipleStorage3.cfs";
    ///
    ///    FileInfo srcFile = new FileInfo(FILENAME);
    ///
    ///    File.Copy(FILENAME, "MultipleStorage_Deleted_Compress.cfs", true);
    ///
    ///    CompoundFile cf = new CompoundFile("MultipleStorage_Deleted_Compress.cfs", ECFUpdateMode.Update, true, true);
    ///
    ///    CFStorage st = cf.RootStorage.GetStorage("MyStorage");
    ///    st = st.GetStorage("AnotherStorage");
    ///    
    ///    Assert.IsNotNull(st);
    ///    st.Delete("Another2Stream"); //17Kb
    ///    cf.Commit();
    ///    cf.Close();
    ///
    ///    CompoundFile.ShrinkCompoundFile("MultipleStorage_Deleted_Compress.cfs");
    ///
    ///    FileInfo dstFile = new FileInfo("MultipleStorage_Deleted_Compress.cfs");
    ///
    ///    Assert.IsTrue(srcFile.Length > dstFile.Length);
    ///
    /// </code>
    /// </example>
    static void ShrinkCompoundFile(std::wstring fileName)
    {
		CFileStream fs;
		fs.Open(fileName.c_str(), L"rb+");
        ShrinkCompoundFile(&fs);
        fs.Close();
    }
};
