/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FSectorCollection.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFList.h"
#include "FSector.h"
#include <vector>
#pragma warning(disable: 4018)

class CFCompoundFile;
typedef void(CFCompoundFile::*Ver3SizeLimitReached)();

class CFSectorArray
{
private:
	std::vector<CFSectorPtr> _arraySectors;

public:
	CFSectorArray(){}
	CFSectorArray(const CFSectorArray& src)
	{
		*this = src;
	}
	~CFSectorArray()
	{
		Clear();
	};

	const CFSector* operator [] (int idx) const
	{
		if (idx<0 || idx>=_arraySectors.size()) return NULL;
		return _arraySectors[idx].c_sector();
	}

	CFSectorPtr& operator [] (int idx)
	{
		if (idx<0) throw CFException("vector subscript out of range"); 

		CFSectorPtr sp;
		if (idx >= _arraySectors.size()) 
			_arraySectors.insert(_arraySectors.cend(),idx-_arraySectors.size()+1, sp);

		return _arraySectors[idx];
	}

	void Add(CFSector* item)
	{
		CFSectorPtr sp(item);
		_arraySectors.push_back(sp);
	}

	void Add(CFSectorPtr item)
	{
		_arraySectors.push_back(item);
	}

	CFSectorArray& operator=(const CFSectorArray& src)
	{
		Clear();
		if (src._arraySectors.size()>0) 
			_arraySectors.assign(src._arraySectors.cbegin(), src._arraySectors.cend());
		return *this;
	}

	void Clear()
	{
		_arraySectors.clear();
	}

	size_t Size()const {return _arraySectors.size();}
};

class CFSectorCollection : public ITList<CFSectorPtr>
{
public:
	CFSectorCollection(void);
	~CFSectorCollection(void);

    static int MAX_SECTOR_V4_COUNT_LOCK_RANGE; // = 524287; //0x7FFFFF00 for Version 4
    static int SLICE_SIZE; // = 4096;

    //public event Ver3SizeLimitReached OnVer3SizeLimitReached;
	Ver3SizeLimitReached _OnVer3SizeLimitReached;
	CFCompoundFile* _compoundFile;
	void AttachCallback(CFCompoundFile* pCompound, Ver3SizeLimitReached pFunc);

private:
	int _count;// = 0;
	std::vector<CFSectorArray> _largeArraySlices;

    //List<ArrayList> largeArraySlices = new List<ArrayList>();

    bool _sizeLimitReached;// = false;

    void DoCheckSizeLimitReached()
    {
        if (!_sizeLimitReached && ((_count - 1) > MAX_SECTOR_V4_COUNT_LOCK_RANGE))
        {
            if (_compoundFile && _OnVer3SizeLimitReached != NULL)
                (_compoundFile->*_OnVer3SizeLimitReached)();

            _sizeLimitReached = true;
        }
    }

public:

	virtual int IndexOf(const CFSectorPtr& item) const
    {
		return IndexOf(item.c_sector());
    }

	int IndexOf(const CFSector* item) const
    {
        return -1;
    }

    void RemoveAt(int index)
    {
		(*this)[index] = NULL;
    }

    virtual CFSectorPtr& operator [](int index)
    {
        int itemIndex = index / SLICE_SIZE;
        int itemOffset = index % SLICE_SIZE;

        if ((index > -1) && (index < _count))
        {
            return _largeArraySlices[itemIndex][itemOffset];
        }
        else
            throw CFException("Argument(index) out of range");
    }

	void Insert(int index, const CFSectorPtr& item) {};

	void SetAt(int index, CFSector* item)
    {
        int itemIndex = index / SLICE_SIZE;
        int itemOffset = index % SLICE_SIZE;

        if (index > -1 && index < _count)
        {
            _largeArraySlices[itemIndex][itemOffset] = item;
        }
        else
            throw CFException("Argument(index) out of range");
    }

private:
	int _Add(CFSector* item)
    {
        int itemIndex = _count / SLICE_SIZE;

		if (itemIndex < _largeArraySlices.size())
        {
            _largeArraySlices[itemIndex].Add(item);
            _count++;
        }
        else
        {
			CFSectorArray sa;
			_largeArraySlices.push_back(sa);
			_largeArraySlices.back().Add(item);
            _count++;
        }

        return _count - 1;
    }

public:
	void Add(CFSector* item)
    {
        DoCheckSizeLimitReached();

        _Add(item);

    }

	virtual void Add(CFSectorPtr& item)
	{
        DoCheckSizeLimitReached();

		_Add(item.sector());
	}


    void Clear()
    {
		std::vector<CFSectorArray>::iterator pos;

        for (pos = _largeArraySlices.begin(); pos != _largeArraySlices.end(); pos++)
        {
            pos->Clear();
        }

        _largeArraySlices.clear();

        _count = 0;
    }
    bool Contains(const CFSectorPtr& item) const
	{
		return IndexOf(item.sector()) >= 0;
	}

    bool Contains(const CFSector* item) const
    {
		return IndexOf(item) >= 0;
    }

    void CopyTo(CFSectorPtr arrayBuf[], int arrayIndex) const
    {
        throw NotImplementedException();
    }

    virtual int Count() const
    {
        return _count;
    }

    bool IsReadOnly() const
    {
        return false;
    }

    bool Remove(const CFSectorPtr& item)
    {
        throw NotImplementedException();
    }

	class SectorEnumerator : public ITEnumerator<CFSector>
	{
	private:
		CFSector* _current;
		std::vector<CFSectorArray>* _array2d;
		int _i;
		int _j;
	public:
		SectorEnumerator()
		{
			_current = NULL;
			_array2d = NULL;
		}
		~SectorEnumerator()
		{
		}

		void SetCollection(std::vector<CFSectorArray>* pArray) { _array2d = pArray;};

		virtual CFSector* CurrentT() { return _current; };
		virtual IFObject* Current() { return (IFObject*)CurrentT(); };

		virtual bool MoveNext() 
		{
			if (_i>=_array2d->size()) return false;

			if (_j>=(*_array2d)[_i].Size())
			{
				_i++; 
				_j = 0;
				if (_i>=_array2d->size()) return false;
			}

			_current = (*_array2d)[_i][_j].sector();
			_j++;

			return true;
		};

		virtual void Reset()
		{
			_i = 0;
			_j = 0;
			_current = NULL;
		};
	};


    virtual SectorEnumerator* GetEnumerator()
    {
		return &_sectorEnumerator;
    }

	virtual ITEnumerator<CFSectorPtr>* GetEnumeratorT() {return NULL;}

protected:
	SectorEnumerator _sectorEnumerator;

};