/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FStorage.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "FileStream.h"
#include "FItem.h"
#include "RBT/BTSearchTree.h"
#include "FStream.h"
#include <list>

//delegate void VisitedEntryAction(CFItem item);

class LIBCDF_EXPORT CFStorage : public CFItem
{
private:
	BTSearchTree<CFItem>* _children;

	CFStorage(void);
public:
	virtual ~CFStorage(void);

    BTSearchTree<CFItem>* Children();

    /// <summary>
    /// Create a new CFStorage
    /// </summary>
    /// <param name="compFile">The Storage Owner - CompoundFile</param>
    CFStorage(CFCompoundFile* compFile);	// only use this ctor

    /// <summary>
    /// Create a CFStorage using an existing directory (previously loaded).
    /// </summary>
    /// <param name="compFile">The Storage Owner - CompoundFile</param>
    /// <param name="dirEntry">An existing Directory Entry</param>
    CFStorage(CFCompoundFile* compFile, IDirectoryEntry* dirEntry);

private:
	BTSearchTree<CFItem>* LoadChildren(int SID);

public:
    /// Check if the storage contains a specified item(either stream, or storage) by a given name
    /// return true if such an item exists
    /// otherwise, returns false
    bool Contains(const std::wstring& itemName);
    
    /// <summary>
    /// Create a new child stream inside the current <see cref="T:OpenMcdf.CFStorage">storage</see>
    /// </summary>
    /// <param name="streamName">The new stream name</param>
    /// <returns>The new <see cref="T:OpenMcdf.CFStream">stream</see> reference</returns>
    /// <exception cref="T:OpenMcdf.CFDuplicatedItemException">Raised when adding an item with the same name of an existing one</exception>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised when adding a stream to a closed compound file</exception>
    /// <exception cref="T:OpenMcdf.CFException">Raised when adding a stream with null or empty name</exception>
    /// <example>
    /// <code>
    /// 
    ///  String filename = "A_NEW_COMPOUND_FILE_YOU_CAN_WRITE_TO.cfs";
    ///
    ///  CompoundFile cf = new CompoundFile();
    ///
    ///  CFStorage st = cf.RootStorage.AddStorage("MyStorage");
    ///  CFStream sm = st.AddStream("MyStream");
    ///  byte[] b = Helpers.GetBuffer(220, 0x0A);
    ///  sm.SetData(b);
    ///
    ///  cf.Save(filename);
    ///  
    /// </code>
    /// </example>
    CFStream* AddStream(const std::wstring& streamName);

    /// <summary>
    /// Get a named <see cref="T:OpenMcdf.CFStream">stream</see> contained in the current storage if existing.
    /// </summary>
    /// <param name="streamName">Name of the stream to look for</param>
    /// <returns>A stream reference if existing</returns>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised if trying to delete item from a closed compound file</exception>
    /// <exception cref="T:OpenMcdf.CFItemNotFound">Raised if item to delete is not found</exception>
    /// <example>
    /// <code>
    /// String filename = "report.xls";
    ///
    /// CompoundFile cf = new CompoundFile(filename);
    /// CFStream foundStream = cf.RootStorage.GetStream("Workbook");
    ///
    /// byte[] temp = foundStream.GetData();
    ///
    /// Assert.IsNotNull(temp);
    ///
    /// cf.Close();
    /// </code>
    /// </example>
    CFStream* GetStream(const std::wstring&streamName);

    /// <summary>
    /// Get a named storage contained in the current one if existing.
    /// </summary>
    /// <param name="storageName">Name of the storage to look for</param>
    /// <returns>A storage reference if existing.</returns>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised if trying to delete item from a closed compound file</exception>
    /// <exception cref="T:OpenMcdf.CFItemNotFound">Raised if item to delete is not found</exception>
    /// <example>
    /// <code>
    /// 
    /// String FILENAME = "MultipleStorage2.cfs";
    /// CompoundFile cf = new CompoundFile(FILENAME, UpdateMode.ReadOnly, false, false);
    ///
    /// CFStorage st = cf.RootStorage.GetStorage("MyStorage");
    ///
    /// Assert.IsNotNull(st);
    /// cf.Close();
    /// </code>
    /// </example>
    CFStorage* GetStorage(const std::wstring& storageName);

    /// <summary>
    /// Create new child storage directory inside the current storage.
    /// </summary>
    /// <param name="storageName">The new storage name</param>
    /// <returns>Reference to the new <see cref="T:OpenMcdf.CFStorage">storage</see></returns>
    /// <exception cref="T:OpenMcdf.CFDuplicatedItemException">Raised when adding an item with the same name of an existing one</exception>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised when adding a storage to a closed compound file</exception>
    /// <exception cref="T:OpenMcdf.CFException">Raised when adding a storage with null or empty name</exception>
    /// <example>
    /// <code>
    /// 
    ///  String filename = "A_NEW_COMPOUND_FILE_YOU_CAN_WRITE_TO.cfs";
    ///
    ///  CompoundFile cf = new CompoundFile();
    ///
    ///  CFStorage st = cf.RootStorage.AddStorage("MyStorage");
    ///  CFStream sm = st.AddStream("MyStream");
    ///  byte[] b = Helpers.GetBuffer(220, 0x0A);
    ///  sm.SetData(b);
    ///
    ///  cf.Save(filename);
    ///  
    /// </code>
    /// </example>
    CFStorage* AddStorage(const std::wstring& storageName);

	//public: List<ICFObject> GetSubTreeObjects()
    //{
    //    List<ICFObject> result = new List<ICFObject>();

    //    _children.VisitTree(TraversalMethod.Inorder,
    //         delegate(BinaryTreeNode<ICFObject> node)
    //         {
    //             result.Add(node.Value);
    //         });

    //    return result;
    //}
    
public:
    void Commit();

private:
	//NodeAction<CFItem>* internalAction;

public: 
    /// <summary>
    /// Visit all entities contained in the storage applying a user provided action
    /// </summary>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised when visiting items of a closed compound file</exception>
    /// <param name="action">User <see cref="T:OpenMcdf.VisitedEntryAction">action</see> to apply to visited entities</param>
    /// <param name="recursive"> Visiting recursion level. True means substorages are visited recursively, false indicates that only the direct _children of this storage are visited</param>
    /// <example>
    /// <code>
    /// const String STORAGE_NAME = "report.xls";
    /// CompoundFile cf = new CompoundFile(STORAGE_NAME);
    ///
    /// FileStream output = new FileStream("LogEntries.txt", FileMode.Create);
    /// TextWriter tw = new StreamWriter(output);
    ///
    /// VisitedEntryAction va = delegate(CFItem item)
    /// {
    ///     tw.WriteLine(item.Name);
    /// };
    ///
    /// cf.RootStorage.VisitEntries(va, true);
    ///
    /// tw.Close();
    /// </code>
    /// </example>
	template <typename _Ty>	// should be a lamda expression or a function pointer
    void VisitEntries(const _Ty& action, bool recursive)
    {
        CheckDisposed();

        //if (action != NULL)
        //{
			typedef BinaryTrees::BinaryTreeNode<CFItem>	BTNode;
			std::list<BTNode*> subStorages;

            //internalAction =
            //    delegate(BinaryTreeNode<CFItem> targetNode)
            //    {
            //        action(targetNode.Value as CFItem);

            //        if (targetNode.Value.DirEntry.Child != DirectoryEntry.NOSTREAM)
            //            subStorages.Add(targetNode);

            //        return;
            //    };

            Children()->VisitTreeInOrder([&](BTNode* targetNode)->bool
			{
                action(targetNode->Value());

                if (targetNode->Value()->DirEntry()->Child() != NOSTREAM)
                    subStorages.push_back(targetNode);
				return true;
			});

            if (recursive && subStorages.size() > 0)
			{
				std::list<BTNode*>::iterator pos = subStorages.begin();
                for (; pos != subStorages.end(); pos++)
                {
					CFStorage* ps = dynamic_cast<CFStorage*>((*pos)->Value());
                    ps->VisitEntries(action, recursive);
                }
			}
        //}
    }


    //public void DeleteStream(String name)
    //{
    //    DeleteItem(name, typeof(CFStream));
    //}

    //public void DeleteStorage(String name)
    //{
    //    DeleteItem(name, typeof(CFStorage));
    //}

public:
    void CopyFrom(CFStorage& src);

public:
    /// <summary>
    /// Remove an entry from the current storage and compound file.
    /// </summary>
    /// <param name="entryName">The name of the entry in the current storage to delete</param>
    /// <example>
    /// <code>
    /// cf = new CompoundFile("A_FILE_YOU_CAN_CHANGE.cfs", UpdateMode.Update, true, false);
    /// cf.RootStorage.DeleteItem("AStream"); // AStream item is assumed to exist.
    /// cf.Commit(true);
    /// cf.Close();
    /// </code>
    /// </example>
    /// <exception cref="T:OpenMcdf.CFDisposedException">Raised if trying to delete item from a closed compound file</exception>
    /// <exception cref="T:OpenMcdf.CFItemNotFound">Raised if item to delete is not found</exception>
    /// <exception cref="T:OpenMcdf.CFException">Raised if trying to delete root storage</exception>
    ///
    /// return false if the element doesn't exist
	bool DeleteItem(const std::wstring& entryName);
    
    /// returns:
    /// 0: the element is renamed
    /// -1: an element with the same name already exists
    /// 1: the element doesn't exist
    int RenameItem(const std::wstring& curName, const std::wstring& newName);

	virtual void Delete()
	{
		delete this;			
	};

};

