/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  FileStream.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//

#pragma once
#include "IFs/IFObject.h"
#include "FException.h"
#include <sys/stat.h>
#include <wchar.h>
#include <stdio.h>
#include <stdlib.h>
#include "time.h"
#include "LibCDF.h"

#pragma warning(disable: 4293)
#pragma warning(disable: 4251)

#ifdef WIN32
typedef long long DLONG;
typedef unsigned long long UDLONG;
#else
typedef long DLONG;
typedef unsigned long UDLONG;
#endif

typedef struct {
	unsigned char byte[4];
}BYTE4;
typedef struct {
	unsigned char byte[8];
}BYTE8;
typedef struct {
	unsigned char byte[16];
}BYTE16;
typedef struct {
	unsigned char byte[64];
}BYTE64;


typedef struct _tagSYSTEMTIME {
    unsigned short wYear;	// base 1900
    unsigned short wMonth;	// base 1
    unsigned short wDayOfWeek;
    unsigned short wDay;
    unsigned short wHour;
    unsigned short wMinute;
    unsigned short wSecond;
    unsigned short wMilliseconds;
} S_SYSTEMTIME, *P_SYSTEMTIME, *LP_SYSTEMTIME;

typedef struct _tagFILETIME {
    unsigned int dwLowDateTime;
    unsigned int dwHighDateTime;
public:
	UDLONG Time64() 
	{
		return (UDLONG)(dwLowDateTime | (UDLONG)dwHighDateTime<<32);
	}
} S_FILETIME, *P_FILETIME, *LP_FILETIME;

#define _MIN(a,b)            (((a) < (b)) ? (a) : (b))
#define _MAX(a,b)            (((a) > (b)) ? (a) : (b))

// Convert Win SystemTime to tm structure.
void SysTime2Tm(const S_SYSTEMTIME * pSystemTime,struct tm & xtm);

class LIBCDF_EXPORT CFDateTime
{
private:
	S_SYSTEMTIME _time;
public:
	CFDateTime()
	{}
	CFDateTime (struct tm* ptm)
	{
		_time.wYear = ptm->tm_year + 1900;
		_time.wMonth = ptm->tm_mon + 1;
		_time.wDayOfWeek = ptm->tm_wday;
		_time.wDay = ptm->tm_mday;
		_time.wHour = ptm->tm_hour;
		_time.wMinute = ptm->tm_min;
		_time.wSecond = ptm->tm_sec;		
	};

	static CFDateTime Now() 
	{
		time_t timep;
		struct tm *p = NULL;
		time(&timep);
		p = localtime(&timep);
		return CFDateTime(p);
	}

	void FromBytes(unsigned char bytes[8]){};

	const S_SYSTEMTIME* GetSysTime() const {return &_time;};

};


class LIBCDF_EXPORT CFileStream : public IFDisposable
{
private:
	std::string _path;
	UDLONG _length;

	bool _canRead;
	bool _canSeek;
	bool _canWrite;

	FILE* _file;

	void Init();
public:
	CFileStream(void);
	CFileStream(const char* filePath, const char* fileMode);
	CFileStream(const wchar_t* filePath, const wchar_t* fileMode);
	~CFileStream(void);

	IMP_PROPERTY(bool, CanRead, _canRead);
	IMP_PROPERTY(bool, CanSeek, _canSeek);
	IMP_PROPERTY(bool, CanWrite, _canWrite);

	bool IsReady() {return (_file != NULL);}

	bool Open(const char* filePath, const char* fileMode)
	{
		Close();
		FILE* p = fopen(filePath,fileMode);
		if (p)
		{
			_path = filePath;
			_file = p;
			return true;
		}
		return false;
	}
	bool Open(const wchar_t* filePath, const wchar_t* fileMode)
	{
		Close();
        char* path = WCS2MBS(filePath);
        char* mode = WCS2MBS(fileMode);
		bool f = Open(path, mode);
        free(path);
        free(mode);
        
        return f;
	}

	virtual UDLONG Length() 
	{
		if (_length > 0) return _length;

		if(_path.length() > 0)
		{
			struct stat statbuff;  
			if(stat(_path.c_str(), &statbuff)>=0)
			{
				_length = (DLONG)statbuff.st_size;
				return _length;
			}
		}
		return -1;
	}

    virtual long Position() const
	{ 
		if (!_file) return -1;
		return ::ftell(_file);
	}
	virtual void Position(long pos, int origin)
	{
		if (_file) ::fseek(_file, pos, origin);
	}

    virtual void Close()
	{
		if (_file) fclose(_file);
		Init();
	}

	typedef enum _tagSeekOrigin
	{
		eBegin,
		eCurrent,
		eEnd
	}ESeekOrigin;

    virtual int Seek(long offset, int origin)
	{
		if (_file) return fseek(_file, offset, origin);
		return -1;
	}

    virtual void SetLength(UDLONG value)
	{
		// we need implement it on Mac: shrink or extend a file
	}

	virtual void Dispose() {Close();};

public:
	virtual void Flush()
	{
		if (_file) fflush(_file);
	};

    virtual int Read(unsigned char buffer[], int offset, int count)
	{
		if (!_file) return -1;

		return (int)fread((void*)(buffer+offset), 1, count, _file); 
	}

    virtual int ReadByte()
	{
		if (!_file) return -1;

		unsigned char pc = 0;
		fread(&pc, 1, 1, _file);
		return (int)pc;
	}

    virtual void Write(const unsigned char* buffer, int offset, int count)
	{
		if (_file)
		{
			fwrite((void*)(buffer+offset), 1, count, _file);
		}
	}

    virtual void WriteByte(unsigned char value)
	{
		fputc(value, _file);
	}
};

