/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  MemoryStream.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//

#pragma once
#include "FileStream.h"
#include <map>

class LIBCDF_EXPORT CFByteArray
{
private:
	unsigned char* _byte;
	unsigned int _size;
    unsigned int _capacity;

	void Init(unsigned int size);


public:
	CFByteArray() : _size(0), _byte(nullptr), _capacity(0) { }
	CFByteArray(unsigned int size)
	{
		Init(size);
	}

	~CFByteArray();

	CFByteArray(int* psValue)
	{
		int ns = sizeof(int);
		Init(ns);
		memcpy(_byte, psValue, ns); //LE order
	}

	unsigned char* Byte()
	{
		return _byte;
	}
	unsigned char* CByte() const 
	{
		return _byte;
	}
	
    unsigned int Length() const {return _size;};
    void SetLength(unsigned int nLen);  //  Set the new length of the buffer.
                                        //  If nLen > Length() then fill dummy data
    
	unsigned int Capacity() const {return _capacity;};
	void Capacity(unsigned int size)
	{
        Clear();
		Init(size);
	}
    
    /// Write data to this buffer, will automatically append the length if needed
    void Write(unsigned int offset, const void* pBuf, unsigned int count);
    /// Append the data to the end
    void Append(const void* pBuf, unsigned int count);

	CFByteArray& operator=(unsigned char data[])
	{
		if (_byte && data) memcpy(_byte, data, _size);
		return *this;
	}
    
	CFByteArray& operator= (const CFByteArray& src);

	void Clear()
	{
		if (_byte != nullptr) free(_byte);
		_byte = NULL;

		_size = 0;
        _capacity = 0;
	}

	void ZeroData()
	{
		if (_byte) memset(_byte, 0, _capacity);
	}

private:
    /// resize the capacity
    void Resize(unsigned int newCap);
};

class CFBuffer
{
public:
	static void BlockCopy(const CFByteArray& src, int srcOffset, CFByteArray& dst, int dstOffset, int count);
	static void BlockCopy(const CFByteArray& src, int srcOffset, unsigned char* dst, int dstOffset, int count);
	static void BlockCopy(const unsigned char* src, int srcOffset, CFByteArray& dst, int dstOffset, int count);
	static void BlockCopy(const unsigned char* src, int srcOffset, unsigned char* dst, int dstOffset, int count);
    static int ByteLength(const CFByteArray& _array);
    static unsigned char GetByte(CFByteArray& _array, int index);
    static void SetByte(CFByteArray& _array, int index, unsigned char value);
};

class CMemoryStream : public CFileStream
{
private:
	CFByteArray _buffer;

	unsigned int _position;
    
public:

    // Summary:
    //     Initializes a new instance of the System.IO.MemoryStream class with an expandable
    //     capacity initialized to zero.
	CMemoryStream();

	~CMemoryStream(void);

    //
    // Summary:
    //     Initializes a new instance of the System.IO.MemoryStream class with an expandable
    //     capacity initialized as specified.
    //
    // Parameters:
    //   capacity:
    //     The initial size of the internal array in bytes.
    //
    // Exceptions:
    //   System.ArgumentOutOfRangeException:
    //     capacity is negative.
	CMemoryStream(int capacity) 
	{
		_position = 0;
		if(capacity>0) _buffer.Capacity((unsigned int)capacity);
	};

public:
    // Summary:
    //     Gets a value indicating whether the current stream supports reading.
    //
    // Returns:
    //     true if the stream is open.
    virtual bool CanRead() const { return true; }
    //
    // Summary:
    //     Gets a value indicating whether the current stream supports seeking.
    //
    // Returns:
    //     true if the stream is open.
    virtual bool CanSeek() const { return true; }
    //
    // Summary:
    //     Gets a value indicating whether the current stream supports writing.
    //
    // Returns:
    //     true if the stream supports writing; otherwise, false.
    virtual bool CanWrite() const { return true; }
    //
    // Summary:
    //     Gets or sets the number of bytes allocated for this stream.
    //
    // Returns:
    //     The length of the usable portion of the buffer for the stream.
    //
    // Exceptions:
    //   System.ArgumentOutOfRangeException:
    //     A capacity is set that is negative or less than the current length of the
    //     stream.
    //
    //   System.ObjectDisposedException:
    //     The current stream is closed.
    //
    //   System.NotSupportedException:
    //     set is invoked on a stream whose capacity cannot be modified.
	virtual unsigned int Capacity() const { return _buffer.Capacity(); }
	virtual void Capacity(unsigned int capacity) { _buffer.Capacity(capacity); }

    //
    // Summary:
    //     Gets the length of the stream in bytes.
    //
    // Returns:
    //     The length of the stream in bytes.
    //
    // Exceptions:
    //   System.ObjectDisposedException:
    //     The stream is closed.
	virtual UDLONG Length() { return (UDLONG)_buffer.Length(); }

	//
    // Summary:
    //     Gets or sets the current position within the stream.
    //
    // Returns:
    //     The current position within the stream.
    //
    // Exceptions:
    //   System.ArgumentOutOfRangeException:
    //     The position is set to a negative value or a value greater than System.Int32.MaxValue.
    //
    //   System.ObjectDisposedException:
    //     The stream is closed.
    virtual long Position() const 
	{
		return (long)_position;
	}
	virtual void Position(long pos, int origin)
	{
		Seek(pos, origin);
	}

	virtual void Close() {};


    // Summary:
    //     Releases the unmanaged resources used by the System.IO.MemoryStream class
    //     and optionally releases the managed resources.
    //
    // Parameters:
    //   disposing:
    //     true to release both managed and unmanaged resources; false to release only
    //     unmanaged resources.
protected: 
	virtual void Dispose(bool disposing) {};
    //
    // Summary:
    //     Overrides System.IO.Stream.Flush() so that no action is performed.
public:
	virtual void Flush() {};
    //
    // Summary:
    //     Returns the array of unsigned bytes from which this stream was created.
    //
    // Returns:
    //     The byte array from which this stream was created, or the underlying array
    //     if a byte array was not provided to the System.IO.MemoryStream constructor
    //     during construction of the current instance.
    //
    // Exceptions:
    //   System.UnauthorizedAccessException:
    //     The MemoryStream instance was not created with a publicly visible buffer.
	virtual CFByteArray& GetBuffer() {return _buffer;};
    //
    // Summary:
    //     Reads a block of bytes from the current stream and writes the data to buffer.
    //
    // Parameters:
    //   buffer:
    //     When this method returns, contains the specified byte array with the values
    //     between offset and (offset + count - 1) replaced by the characters read from
    //     the current stream.
    //
    //   offset:
    //     The byte offset in buffer at which to begin reading.
    //
    //   count:
    //     The maximum number of bytes to read.
    //
    // Returns:
    //     The total number of bytes written into the buffer. This can be less than
    //     the number of bytes requested if that number of bytes are not currently available,
    //     or zero if the end of the stream is reached before any bytes are read.
    //
    // Exceptions:
    //   System.ArgumentNullException:
    //     buffer is null.
    //
    //   System.ArgumentOutOfRangeException:
    //     offset or count is negative.
    //
    //   System.ArgumentException:
    //     offset subtracted from the buffer length is less than count.
    //
    //   System.ObjectDisposedException:
    //     The current stream instance is closed.
    virtual int Read(CFByteArray& buffer, int offset, int count);

	virtual int Read(unsigned char buffer[], int offset, int count);

    //
    // Summary:
    //     Reads a byte from the current stream.
    //
    // Returns:
    //     The byte cast to a System.Int32, or -1 if the end of the stream has been
    //     reached.
    //
    // Exceptions:
    //   System.ObjectDisposedException:
    //     The current stream instance is closed.
    virtual int ReadByte();
    //
    // Summary:
    //     Sets the position within the current stream to the specified value.
    //
    // Parameters:
    //   offset:
    //     The new position within the stream. This is relative to the loc parameter,
    //     and can be positive or negative.
    //
    //   loc:
    //     A value of type System.IO.SeekOrigin, which acts as the seek reference point.
    //
    // Returns:
    //     The new position within the stream, calculated by combining the initial reference
    //     point and the offset.
    //
    // Exceptions:
    //   System.IO.IOException:
    //     Seeking is attempted before the beginning of the stream.
    //
    //   System.ArgumentOutOfRangeException:
    //     offset is greater than System.Int32.MaxValue.
    //
    //   System.ArgumentException:
    //     There is an invalid System.IO.SeekOrigin. -or- offset caused an arithmetic
    //     overflow.
    //
    //   System.ObjectDisposedException:
    //     The current stream instance is closed.
	virtual int Seek(long offset, int origin);
    //
    // Summary:
    //     Sets the length of the current stream to the specified value.
    //
    // Parameters:
    //   value:
    //     The value at which to set the length.
    //
    // Exceptions:
    //   System.NotSupportedException:
    //     The current stream is not resizable and value is larger than the current
    //     capacity.  -or- The current stream does not support writing.
    //
    //   System.ArgumentOutOfRangeException:
    //     value is negative or is greater than the maximum length of the System.IO.MemoryStream,
    //     where the maximum length is(System.Int32.MaxValue - origin), and origin is
    //     the index into the underlying buffer at which the stream starts.
    virtual void SetLength(UDLONG value);

	//
    // Summary:
    //     Writes a block of bytes to the current stream using data read from buffer.
    //
    // Parameters:
    //   buffer:
    //     The buffer to write data from.
    //
    //   offset:
    //     The byte offset in buffer at which to begin writing from.
    //
    //   count:
    //     The maximum number of bytes to write.
    //
    // Exceptions:
    //   System.ArgumentNullException:
    //     buffer is null.
    //
    //   System.NotSupportedException:
    //     The stream does not support writing. For additional information see System.IO.Stream.CanWrite.
    //      -or- The current position is closer than count bytes to the end of the stream,
    //     and the capacity cannot be modified.
    //
    //   System.ArgumentException:
    //     offset subtracted from the buffer length is less than count.
    //
    //   System.ArgumentOutOfRangeException:
    //     offset or count are negative.
    //
    //   System.IO.IOException:
    //     An I/O error occurs.
    //
    //   System.ObjectDisposedException:
    //     The current stream instance is closed.
    virtual void Write(CFByteArray& buffer, int offset, int count);
    virtual void Write(unsigned char buffer[], int offset, int count);

	//
    // Summary:
    //     Writes a byte to the current stream at the current position.
    //
    // Parameters:
    //   value:
    //     The byte to write.
    //
    // Exceptions:
    //   System.NotSupportedException:
    //     The stream does not support writing. For additional information see System.IO.Stream.CanWrite.
    //      -or- The current position is at the end of the stream, and the capacity
    //     cannot be modified.
    //
    //   System.ObjectDisposedException:
    //     The current stream is closed.
    virtual void WriteByte(unsigned char value);
    //
    // Summary:
    //     Writes the entire contents of this memory stream to another stream.
    //
    // Parameters:
    //   stream:
    //     The stream to write this memory stream to.
    //
    // Exceptions:
    //   System.ArgumentNullException:
    //     stream is null.
    //
    //   System.ObjectDisposedException:
    //     The current or target stream is closed.
    virtual void WriteTo(CFileStream* stream);

};

