/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  BTNode.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include <vector>

namespace BinaryTrees
{
	template<class _T> class Node;

    /// <summary>
    /// Represents a collection of Node<_T> instances.
    /// </summary>
    /// <typeparam name="T">The type of data held in the Node instances referenced by this class.</typeparam>
	template<class _T>
    class NodeList : public std::vector< Node<_T>* >, public RefObject
    {
		typedef Node<_T> _BaseNode;
		typedef std::vector< Node<_T>* > _BaseClass;
	public:
		NodeList() : _BaseClass(), RefObject() { }

        NodeList(int initialSize, bool newNode, _BaseNode* pBase)
        {
			if(newNode)
			{
				// Add the specified number of items
				for (int i = 0; i < initialSize; i++)
					_BaseClass::push_back(pBase->NewNode());
			}
			else
			{
                _BaseClass::resize(initialSize);
			}
        }

		virtual ~NodeList()
		{
			typename _BaseClass::iterator pos = _BaseClass::begin();
			for(; pos != _BaseClass::end(); pos++)
			{
				if (!(*pos)) continue;
				delete (*pos);
			}
			_BaseClass::clear();		
		}

        virtual bool Equals(const IFObject* obj) const
		{
			return false;
		};

		virtual void Delete() 
		{
			delete this;
		}

		bool IsRefed()
		{
			typename _BaseClass::iterator pos = _BaseClass::begin();
			for(; pos != _BaseClass::end(); pos++)
			{
				if (!(*pos)) continue;
				if((*pos)->Value()) return true;
			}

			return false;
		}

        /// <summary>
        /// Searches the NodeList for a Node containing a particular value.
        /// </summary>
        /// <param name="value">The value to search for.</param>
        /// <returns>The Node in the NodeList, if it exists; null otherwise.</returns>
        Node<_T>* FindByValue(_T* value)
        {
			if (!value) return NULL;

            // search the list for the value
			typename _BaseClass::const_iterator cpos = _BaseClass::cbegin();
			for(; cpos != _BaseClass::cend(); cpos++){
				if (!(*cpos)) continue;
                if (value->Equals((*cpos)->Value()))
                    return (*cpos);
			}

            // if we reached here, we didn't find a matching node
            return NULL;
        }

		void SetAt(int idx, Node<_T>* pNode)
		{
			if (idx<0 || idx >= _BaseClass::size()) return ;
			if (pNode && (*this)[idx]) 
			{
				throw CFException("This item is not NULL.");
			}
			(*this)[idx] = pNode;
		}
		void SetAt(int idx, _T* data, NodeList<_T>* neighbors)
		{
			if (idx<0 || idx >= _BaseClass::size()) return ;
			Node<_T>* pNode = (*this)[idx];
			if (pNode)
			{
				pNode->Value(data);
				pNode->Neighbors(neighbors);
			}
		};

		Node<_T>* GetAt(int idx) const
		{
			if (idx<0 || idx >= _BaseClass::size()) return NULL;
			return (*this)[idx];
		};
    };

	template<class _T>
	class Node
	{
	private:
		_T* _data;
	protected:
		NodeList<_T>* _neighbors;

	public:
		Node() { _data = NULL; _neighbors = NULL;}
		Node(_T* data)
		{
			_data = NULL;
			_neighbors = NULL;
			Value(data);
		}
		Node(_T* data, NodeList<_T>* neighbors)
		{
			_data = NULL;
			_neighbors = NULL;
			Value(data);
			Neighbors(neighbors);
		}

		virtual ~Node()
		{
			if (_data) _data->ReleaseRef();
			if (_neighbors) _neighbors->ReleaseRef();
		}

        _T* Value() const 
        {
            return _data;
        }
        void Value(_T* value)
        {
			if (value) value->AddRef();
			if (_data) _data->ReleaseRef();
            _data = value;
        }

		Node<_T>& operator=(const Node<_T>& src)
		{
			Value(src.Value());
			Neighbors(src.Neighbors());
			return *this;
		}

		virtual Node<_T>* NewNode()
		{
			Node<_T>* p = new Node<_T>();
			return p;
		}

		friend class NodeList<_T>; 
	protected: 
		NodeList<_T>* Neighbors() const
        {
            return _neighbors;
        }
        void Neighbors(NodeList<_T>* list)
        {
			if (_neighbors && _neighbors->IsRefed()) 
			{
				throw CFException("This node list is using.");
			}
			else{
				if (list) list->AddRef();
				if (_neighbors) _neighbors->ReleaseRef();
				_neighbors = list;
			}
        }
   };

	/// <summary>
    /// The BinaryTreeNode class represents a node in a binary tree, or a binary search tree.
    /// It has precisely two neighbors, which can be accessed via the Left and Right properties.
    /// </summary>
    /// <typeparam name="T">The type of data stored in the binary tree node.</typeparam>
	template<class _T>
    class BinaryTreeNode : public Node<_T>
    {
	public:
		typedef Node<_T> _BaseClass;
		typedef BinaryTreeNode<_T> _BTNodeClass;
		BinaryTreeNode() : _BaseClass() {}
        BinaryTreeNode(_T* data) : _BaseClass(data, NULL) {}

		BinaryTreeNode(const Node<_T>& base) 
		{
			Value(base.Value());
			Neighbors(base.Neighbors());
		}

        BinaryTreeNode(_T* data, const BinaryTreeNode<_T>& left, const BinaryTreeNode<_T>& right)
        {
            Value(data);

            NodeList<_T>* children = new NodeList<_T>(2, true, this);
            children->SetAt(0, left.Value(), left.Neighbors());
            children->SetAt(1, right.Value(), right.Neighbors());

            Neighbors(children);
        }

		virtual ~BinaryTreeNode()
		{
		}

		virtual Node<_T>* NewNode()
		{
			Node<_T>* p = new BinaryTreeNode<_T>();
			return p;
		}

		BinaryTreeNode<_T>* CutLeft()
		{
            if (_BaseClass::_neighbors == NULL)
                return NULL;
            else
			{
                BinaryTreeNode<_T>* pNode = dynamic_cast<_BTNodeClass*>(_BaseClass::_neighbors->GetAt(0));
				if (!pNode) return NULL;
				_BaseClass::_neighbors->SetAt(0, NULL);
				if (pNode->Value()) return pNode;
			}
            return NULL;
		}
        BinaryTreeNode<_T>* Left() const
        {
            if (_BaseClass::_neighbors == NULL)
                return NULL;
            else
			{
                BinaryTreeNode<_T>* pNode = dynamic_cast<_BTNodeClass*>(_BaseClass::_neighbors->GetAt(0));
				if (pNode && pNode->Value()) return pNode;
			}
			return NULL;
        }
        void Left(const BinaryTreeNode<_T>& lNode)
        {
            if (_BaseClass::_neighbors == NULL) {
                Neighbors(new NodeList<_T>(2, true, this));
			}

            _BaseClass::_neighbors->SetAt(0, lNode.Value(), lNode.Neighbors());
        }
        void Left(BinaryTreeNode<_T>* lNode)
		{
			if (!lNode)
			{
	            if (_BaseClass::_neighbors == NULL) return;
				else _BaseClass::_neighbors->SetAt(0, NULL, NULL);
			}
			else
			{
				if (!_BaseClass::_neighbors) _BaseClass::Neighbors(new NodeList<_T>(2, false, NULL));
				_BaseClass::_neighbors->SetAt(0, lNode);
			}
		}

		BinaryTreeNode<_T>* CutRight()
		{
            if (_BaseClass::_neighbors == NULL)
                return NULL;
            else
			{
                BinaryTreeNode<_T>* pNode = dynamic_cast<_BTNodeClass*>(_BaseClass::_neighbors->GetAt(1));
				if (!pNode) return NULL;
				_BaseClass::_neighbors->SetAt(1, NULL);
				if (pNode->Value()) return pNode;
			}
            return NULL;
		}

        BinaryTreeNode<_T>* Right() const
        {
            if (_BaseClass::_neighbors == NULL)
                return NULL;
            else
			{
                BinaryTreeNode<_T>* pNode = dynamic_cast<_BTNodeClass*>(_BaseClass::_neighbors->GetAt(1));
				if (pNode && pNode->Value()) return pNode;
			}
            return NULL;
        }
        void Right(const BinaryTreeNode<_T>& rNode)
        {
            if (_BaseClass::_neighbors == NULL)
			{
                Neighbors(new NodeList<_T>(2, true, this));
			}

            _BaseClass::_neighbors->SetAt(1, rNode.Value(), rNode.Neighbors());
        }
        void Right(BinaryTreeNode<_T>* rNode)
		{
			if (!rNode)
			{
	            if (_BaseClass::_neighbors == NULL) return;
				else _BaseClass::_neighbors->SetAt(1, NULL, NULL);
			}
			else
			{
				if (!_BaseClass::_neighbors) _BaseClass::Neighbors(new NodeList<_T>(2, false, NULL));
				_BaseClass::_neighbors->SetAt(1, rNode);
			}
		}
    };
};