/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  BTSearchTree.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "../IFs/IFCollection.h"
#include "../IFs/IFEnumerable.h"
#include "../IFs/IFComparable.h"
#include "../FException.h"
#include "BTNode.h"
#include <exception>
#include <string>
#include <stack>

class BSTDuplicatedException : public CFException
{
public:
	BSTDuplicatedException(void);
	BSTDuplicatedException(const char* message);
	BSTDuplicatedException(const std::string& message);
	virtual ~BSTDuplicatedException(void);
    
public:
	virtual const char * what() const throw();
    
};

typedef enum _tagTraversalMeth
{
    eBTPreorder,
    eBTInorder,
    eBTPostorder
}ETraversalMethod;

template<class _T>
class INodeAction 
{
public:
	typedef BinaryTrees::BinaryTreeNode<_T> BTNode;
	virtual ~INodeAction() {}
	virtual bool operator() (BTNode* node) = 0;
};

template <typename _Ty, class _T>
class NodeAction : public INodeAction<_T>
{
public:
	typedef INodeAction<_T> baseClass;
	NodeAction(const _Ty& action): mAction(action) {}
	virtual bool operator() (typename baseClass::BTNode* node) {
		return mAction(node);
	}

protected:
	_Ty		mAction;
};

template<class _T>
class BTSearchTree : public ITCollectionPtr<_T>
{
public:
	BTSearchTree() 
	{
		_count = 0;
		_root = NULL;
		_comparer = new TObjectComparer<_T>();
		_preEnumer.Tree(this);
		_inEnumer.Tree(this);
		_postEnumer.Tree(this);
	};

	typedef ITComparer<_T> BTComparer;
	typedef BinaryTrees::BinaryTreeNode<_T> BTNode;

    BTSearchTree(BTComparer* comparer)
    {
		_count = 0;
		_root = NULL;
        _comparer = comparer;
		_preEnumer.Tree(this);
		_inEnumer.Tree(this);
		_postEnumer.Tree(this);
    };

	virtual ~BTSearchTree() 
	{
		if (_root) delete _root;
		if (_comparer) delete _comparer;
	}
private:
    int _count;
	BTNode* _root;
    BTComparer* _comparer;

public:
    void Clear()
    {
		if (_root) delete _root;
        _root = NULL;
        _count = 0;
    }

    virtual void CopyTo(_T* arrbuffer[], int index)
    {
        _CopyTo(arrbuffer, index, eBTInorder);
    }

    void _CopyTo(_T* arrBuffer[], int index, ETraversalMethod eTraversalMethod)
    {
        ITEnumerator<_T>* enumProp = NULL;

        switch (eTraversalMethod)
        {
		case eBTPreorder:
                enumProp = PreorderTraverser();
                break;

		case eBTInorder:
                enumProp = InorderTraverser();
                break;

		case eBTPostorder:
		default:
                enumProp = PostorderTraverser();
                break;
        }

        int i = 0;
		while(enumProp->MoveNext())
        {
            arrBuffer[i + index] = enumProp->CurrentT();
            i++;
        }
    }

    virtual void Add(_T* data)
    {
        BTNode* n = new BTNode(data);
        int result;

        BTNode* current = _root, *parent = NULL;
		int cur_result = 0;
        while (current != NULL)
        {
            result = _comparer->Compare(current->Value(), data);
            if (result == 0)
			{
				delete n;
                throw BSTDuplicatedException();
			}
            else if (result < 0)
            {
                parent = current;
                current = current->Right();
				cur_result = -1;
            }
            else if (result > 0)
            {
                parent = current;
                current = current->Left();
				cur_result = 1;
            }
        }

        _count++;
        if (parent == NULL)
            _root = n;
        else
        {
            if (cur_result > 0)
                parent->Left(n);
            else
                parent->Right(n);
        }
    }

    bool Contains(_T* data) const
    {
        BTNode* current = _root;
        int result;
        while (current != NULL)
        {
            result = _comparer->Compare(current->Value(), data);
            if (result == 0)
                return true;
            else if (result > 0)
                current = current->Left();
            else if (result < 0)
                current = current->Right();
        }

        return false;
    }

    bool TryFind(_T* data, _T*& foundObject) 
    {
        foundObject = NULL;

        int result;
        BTNode* current = _root;
        while (current != NULL)
        {
            result = _comparer->Compare(current->Value(), data);
            if (result == 0)
            {
                foundObject = current->Value();
                return true;
            }
            else if (result > 0)
                current = current->Left();
            else if (result < 0)
                current = current->Right();
        }

        return false;
    }

    virtual bool Remove(_T* data)
    {
        if (_root == NULL)
            return false;

        BTNode* current = _root, *parent = NULL;
        int result = _comparer->Compare(current->Value(), data);
		int cur_result = 0;
        while (result != 0)
        {
            if (result > 0)
            {
                parent = current;
                current = current->Left();
				cur_result = 1;
            }
            else if (result < 0)
            {
                parent = current;
                current = current->Right();
				cur_result = -1;
            }

            if (current == NULL)
                return false;
            else
                result = _comparer->Compare(current->Value(), data);
        }

		if(parent)
		{
			if (cur_result>0) parent->CutLeft();	// current
			else parent->CutRight();
		}
        _count--;

        if (current->Right() == NULL)
        {
            if (parent == NULL)
			{
                _root = current->CutLeft();
			}
            else
            {
                if (cur_result > 0)
                    parent->Left(current->CutLeft());
                else if (cur_result < 0)
                    parent->Right(current->CutLeft());
            }
        }
        else if (current->Right()->Left() == NULL)
        {
            current->Right()->Left(current->CutLeft());

            if (parent == NULL)
                _root = current->CutRight();
            else
            {
                if (cur_result > 0)
                    parent->Left(current->CutRight());
                else if (cur_result < 0)
                    parent->Right(current->CutRight());
            }
        }
        else
        {
            BTNode* leftmost = current->Right()->Left(), *lmParent = current->Right();
            while (leftmost->Left() != NULL)
            {
                lmParent = leftmost;
                leftmost = leftmost->Left();
            }
			lmParent->CutLeft(); // leftmost
            lmParent->Left(leftmost->CutRight());

            leftmost->Left(current->CutLeft());
            leftmost->Right(current->CutRight());

            if (parent == NULL)
                _root = leftmost;
            else
            {
                if (cur_result > 0)
                    parent->Left(leftmost);
                else if (cur_result < 0)
                    parent->Right(leftmost);
            }
        }

        current->Left(NULL);
		current->Right(NULL);
		delete current;
        current = NULL;

        return true;
    }
    int Count() const
    {
        return _count;
    }

	bool IsReadOnly() const {return false;}

	virtual ITEnumerator<_T>* GetEnumeratorT() {return _GetEnumerator();}

    ITEnumerator<_T>* _GetEnumerator()
    {
        return GetEnumerator(eBTInorder);
    }

	virtual IFEnumerator* GetEnumerator()
    {
        return GetEnumerator(eBTInorder);
    }

    virtual ITEnumerator<_T>* GetEnumerator(ETraversalMethod eTraversalMethod)
    {
        switch (eTraversalMethod)
        {
            case eBTPreorder:
                return PreorderTraverser();

            case eBTInorder:
                return InorderTraverser();

            case eBTPostorder:
            default:
                return PostorderTraverser();
        }
    }

	template <typename _Ty>
	void VisitTreeInOrder(const _Ty& action) 
	{
		NodeAction<_Ty, _T>* pAction = new NodeAction<_Ty, _T>(action);		
		_VisitTreeInOrder(pAction);
		delete pAction;
	}

	void _VisitTreeInOrder(INodeAction<_T>* nodeCallback)
    {
        if (_root != NULL)
            DoVisitInOrder(_root, nodeCallback);
    }

    void _VisitTreeInOrder(INodeAction<_T>* nodeCallback, bool recursive)
    {
        if (_root != NULL)
            DoVisitInOrder(_root, nodeCallback);
    }

private:
	void DoVisitInOrder(BTNode* node, INodeAction<_T>* nodeCallback)
    {
        if (node->Left() != NULL) DoVisitInOrder(node->Left(), nodeCallback);

        if (nodeCallback != NULL)
        {
            (*nodeCallback)(node);
        }

        if (node->Right() != NULL) DoVisitInOrder(node->Right(), nodeCallback);
    }
public:
	BTNode* Root() const {return _root;}

	class PreorderEnumerator : public ITEnumerator<_T>
	{
	private:
		_T* _current;
		std::stack<BTNode*> _toVisit;
		BTSearchTree<_T>* _pTree;
	public:
		PreorderEnumerator()
		{
			_current = NULL;
			_pTree = NULL;
		}
		~PreorderEnumerator()
		{
		}

		void Tree(BTSearchTree<_T>* tree) {_pTree = tree;}

		virtual _T* CurrentT() { return _current; };
		virtual IFObject* Current() { return (IFObject*)CurrentT(); };

		virtual bool MoveNext() 
		{
			if (_toVisit.size() == 0) return false;

            // take the top item from the stack
			BTNode* currentNode = _toVisit.top(); 
			_toVisit.pop();

			if (!currentNode) return false;
            // add the right and left children, if not NULL
            if (currentNode->Right() != NULL) _toVisit.push(currentNode->Right());
            if (currentNode->Left() != NULL) _toVisit.push(currentNode->Left());

            // return the current node
            _current = currentNode->Value();
			return true;
		};

		virtual void Reset()
		{
			if (_pTree)
			{
				while(_toVisit.size()>0) {_toVisit.pop();}//_toVisit._Get_container().clear();
				_toVisit.push(_pTree->Root());
			}
			_current = NULL;
		};
	};

	ITEnumerator<_T>* PreorderTraverser()
	{
		_preEnumer.Reset();
		return &_preEnumer;
	}

	class InorderEnumerator : public ITEnumerator<_T>
	{
	private:
		_T* _current;
		BTNode* _currentNode;
		std::stack<BTNode*> _toVisit;
		BTSearchTree<_T>* _pTree;
	public:
		InorderEnumerator()
		{
			_current = NULL;
			_currentNode = NULL;
			_pTree = NULL;
		}
		~InorderEnumerator()
		{
		}

		void Tree(BTSearchTree<_T>* tree) {_pTree = tree;}

		virtual _T* CurrentT() { return _current; };
		virtual IFObject* Current() { return (IFObject*)CurrentT(); };

		virtual bool MoveNext() 
		{
			if (!_currentNode && _toVisit.size() == 0) return false;

            // Get the left-most item in the subtree, remembering the path taken
            while (_currentNode != NULL)
            {
                _toVisit.push(_currentNode);
                _currentNode = _currentNode->Left();
            }

            _currentNode = _toVisit.top();
			_toVisit.pop();

			_current = _currentNode->Value();
			_currentNode = _currentNode->Right();
			return true;
		};

		virtual void Reset()
		{
			if (_pTree)
			{
				while(_toVisit.size()>0) {_toVisit.pop();}//_toVisit._Get_container().clear();
				_currentNode = _pTree->Root();
			}
			_current = NULL;
		};
	};

	ITEnumerator<_T>* InorderTraverser()
	{
		_inEnumer.Reset();
		return &_inEnumer;
	}

	class PostorderEnumerator : public ITEnumerator<_T>
	{
	private:
		_T* _current;
		BTNode* _currentNode;
		std::stack<BTNode*> _toVisit;
		std::stack<bool> _hasBeenProcessed;
		BTSearchTree<_T>* _pTree;
	public:
		PostorderEnumerator()
		{
			_current = NULL;
			_currentNode = NULL;
			_pTree = NULL;
		}
		~PostorderEnumerator()
		{
		}

		void Tree(BTSearchTree<_T>* tree) {_pTree = tree;}

		virtual _T* CurrentT() { return _current; };
		virtual IFObject* Current() { return (IFObject*)CurrentT(); };

		virtual bool MoveNext() 
		{
			if (_toVisit.size() == 0) return false;

			while(_toVisit.size()>0)
			{
				if (_currentNode != NULL)
				{
					// add this node to the stack with a false processed value
					_toVisit.push(_currentNode);
					_hasBeenProcessed.push(false);
					_currentNode = _currentNode->Left();
				}
				else
				{
					// see if the node on the stack has been processed
					bool processed = _hasBeenProcessed.top();
					_hasBeenProcessed.pop();
					BinaryTrees::BinaryTreeNode<_T>* node = _toVisit.top();
					_toVisit.pop();
					if (!processed)
					{
						// if it's not been processed, "recurse" down the right subtree
						_toVisit.push(node);
						_hasBeenProcessed.push(true);    // it's now been processed
						_currentNode = node->Right();
					}
					else
					{
						_current = node->Value();
						break;
					}
				}
			}
			return true;
		};

		virtual void Reset()
		{
			if (_pTree)
			{
				while(_toVisit.size()>0) {_toVisit.pop();}//_toVisit._Get_container().clear();
				while(_hasBeenProcessed.size()>0) {_hasBeenProcessed.pop();}//_hasBeenProcessed._Get_container().clear();

				_currentNode = _pTree->Root();
				if (_currentNode != NULL)
				{
					_toVisit.push(_currentNode);
					_hasBeenProcessed.push(false);
					_currentNode = _currentNode->Left();
				}
			}
			_current = NULL;
		};
	};

	ITEnumerator<_T>* PostorderTraverser()
	{
		_postEnumer.Reset(); 
		return &_postEnumer;
	}

private:
	PreorderEnumerator _preEnumer;
	InorderEnumerator _inEnumer;
	PostorderEnumerator _postEnumer;
};

