/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

#pragma once
#include "../IFs/IFComparable.h"
#include "RBTNode.h"
#include "RBTEnumerator.h"
#include <string>
#include "../LibCDF.h"

//namespace RBT
//{
///<summary>
///A red-black tree must satisfy these properties:
///
///1. The root is black. 
///2. All leaves are black. 
///3. Red nodes can only have black children. 
///4. All paths from a node to its leaves contain the same number of black nodes.
///</summary>
class RBTree : public RefObject
{
private:
	// the number of nodes contained in the tree
	int             intCount;
    // a simple randomized hash code. The hash code could be used as a key
    // if it is "unique" enough. Note: The IComparable interface would need to 
    // be replaced with int.
    int             intHashCode;
	// identifies the owner of the tree
	std::string     strIdentifier;
	// the tree
	RBTNode*		rbTree;
    // the node that was last found; used to optimize searches
	RBTNode*		lastNodeFound;			
	CFRandom			rand;

public:
    //  sentinelNode is convenient way of indicating a leaf node.
    static RBTNode* sentinelNode;          

	RBTree(); 
		
	RBTree(std::string strIdentifier); 
		
	///<summary>
	/// Add
	/// args: ByVal key As IComparable, ByVal data As Object
	/// key is object that implements IComparable interface
	/// performance tip: change to use use int type (such as the hashcode)
	///</summary>
	void Add(IFComparable* key, IFObject* data);

private:
	///<summary>
    /// RestoreAfterInsert
    /// Additions to red-black trees usually destroy the red-black 
    /// properties. Examine the tree and restore. Rotations are normally 
    /// required to restore it
    ///</summary>
	void RestoreAfterInsert(RBTNode* x);
	
public:
	///<summary>
	/// RotateLeft
	/// Rebalance the tree by rotating the nodes to the left
	///</summary>
	void RotateLeft(RBTNode* x);
	///<summary>
	/// RotateRight
	/// Rebalance the tree by rotating the nodes to the right
	///</summary>
	void RotateRight(RBTNode* x);
	///<summary>
	/// GetData
	/// Gets the data object associated with the specified key
	///<summary>
	IFObject* GetData(const IFComparable* key);
	///<summary>
	/// GetMinKey
	/// Returns the minimum key value
	///<summary>
	IFComparable* GetMinKey();
	///<summary>
	/// GetMaxKey
	/// Returns the maximum key value
	///<summary>
	IFComparable* GetMaxKey();
	///<summary>
	/// GetMinValue
	/// Returns the object having the minimum key value
	///<summary>
	IFObject* GetMinValue();
	///<summary>
	/// GetMaxValue
	/// Returns the object having the maximum key
	///<summary>
	IFObject* GetMaxValue();
	///<summary>
	/// GetEnumerator
	/// return an enumerator that returns the tree nodes in order
	///<summary>
	RBTEnumerator* GetEnumerator();
	///<summary>
	/// Keys
	/// if(ascending is true, the keys will be returned in ascending order, else
	/// the keys will be returned in descending order.
	///<summary>
	RBTEnumerator* Keys();
	RBTEnumerator* Keys(bool ascending);
	///<summary>
	/// Values
	/// Provided for .NET compatibility. 
	///<summary>
	RBTEnumerator* Values();
	///<summary>
	/// Elements
	/// Returns an enumeration of the data objects.
	/// if(ascending is true, the objects will be returned in ascending order,
	/// else the objects will be returned in descending order.
	///<summary>
	RBTEnumerator* Elements();
	RBTEnumerator* Elements(bool ascending);
	///<summary>
	/// IsEmpty
	/// Is the tree empty?
	///<summary>
	bool IsEmpty() const;
	///<summary>
	/// Remove
	/// removes the key and data object (delete)
	///<summary>
	void Remove(IFComparable* key);
private:
	///<summary>
	/// Delete
	/// Delete a node from the tree and restore red black properties
	///<summary>
	void Delete(RBTNode* z);

    ///<summary>
    /// RestoreAfterDelete
    /// Deletions from red-black trees may destroy the red-black 
    /// properties. Examine the tree and restore. Rotations are normally 
    /// required to restore it
    ///</summary>
	void RestoreAfterDelete(RBTNode* x);
	
public:
	///<summary>
	/// RemoveMin
	/// removes the node with the minimum key
	///<summary>
	void RemoveMin();
	///<summary>
	/// RemoveMax
	/// removes the node with the maximum key
	///<summary>
	void RemoveMax();
	///<summary>
	/// Clear
	/// Empties or clears the tree
	///<summary>
	void Clear ();
	///<summary>
	/// Size
	/// returns the size (number of nodes) in the tree
	///<summary>
	int Size();
	///<summary>
	/// Equals
	///<summary>
	//override 
	virtual bool Equals(const IFObject* obj) const;
	///<summary>
	/// HashCode
	///<summary>
	//override 
	virtual int GetHashCode();
	///<summary>
	/// ToString
	///<summary>
	//override 
	virtual std::string ToString() const;

	virtual void Delete() {delete this;};
};
//}