/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
 */

//
//  FDirectoryEntry.cpp
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#include "StdAfx.h"
#include "FDirectoryEntry.h"
#include "FSector.h"
#include <time.h>

CFDirectoryEntry::CFDirectoryEntry(void)
{
	sid = -1;
    stgType = eStgInvalid;
    stgColor = eBlack;
    leftSibling = NOSTREAM;
    rightSibling = NOSTREAM;
    child = NOSTREAM;
	memset(storageCLSID.byte, 0, 16);	// = Guid.NewGuid();	//need a method to create the new GUID on Mac?
    stateBits = 0;;
	memset(creationDate.byte, 0, 8);	// = new byte[8] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
	memset(modifyDate.byte, 0, 8);		// = new byte[8] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
    startSector = CFSector::ENDOFCHAIN;
    size = 0;
}

CFDirectoryEntry::CFDirectoryEntry(StgType argStgType)
{
	sid = -1;
    stgColor = eBlack;
    leftSibling = NOSTREAM;
    rightSibling = NOSTREAM;
    child = NOSTREAM;
    stateBits = 0;;
    startSector = CFSector::ENDOFCHAIN;
    size = 0;

	stgType = argStgType;
	time_t tCurrent = 0;

    switch (stgType)
    {
        case eStgStream:
			memset(storageCLSID.byte, 0, 16);
			memset(creationDate.byte, 0, 8);
			memset(modifyDate.byte, 0, 8);
            break;

        case eStgStorage:
			memset(storageCLSID.byte, 0, 16);
			time(&tCurrent);	// need to verify under Mac, sine Jan 1 1970 0r Jan 1 1601
			memcpy(creationDate.byte, &tCurrent, sizeof(time_t)); // need to verify the BIG_ENDIAN/LITTLE_ENDIAN
			memset(modifyDate.byte, 0, 8);
            break;

        case eStgRoot:
			memset(storageCLSID.byte, 0, 16);
			memset(creationDate.byte, 0, 8);
			memset(modifyDate.byte, 0, 8);
            break;
        default:
            break;
    }

}

CFDirectoryEntry::~CFDirectoryEntry(void)
{
}

const wchar_t* CFDirectoryEntry::GetEntryName() const
{
    return entryName.c_str();
}

void CFDirectoryEntry::SetEntryName(const std::wstring& enName)
{
    if (enName.find(L'\\') != std::wstring::npos ||
        enName.find(L'/') != std::wstring::npos ||
        enName.find(L':') != std::wstring::npos ||
        enName.find(L'!') != std::wstring::npos)
        throw CFException("Invalid character in entry: the characters '\\', '/', ':','!' cannot be used in entry name");
    
    if (enName.length() > 31)
        throw CFException("Entry name MUST be smaller than 31 characters");
    
    entryName = enName;
}

int CFDirectoryEntry::CompareTo(const IFComparable* obj) const
{
    const CFDirectoryEntry* otherDir = dynamic_cast<const CFDirectoryEntry*>(obj);
    
    if (otherDir == NULL)
        throw CFException("Invalid casting: compared object does not implement IDirectorEntry interface");
    
    if (entryName.length() > otherDir->entryName.length())
    {   return 1; }
    else if (entryName.length() < otherDir->entryName.length())
    {   return -1; }
    else
    {
        const wchar_t* pEntryname = entryName.c_str();
        const wchar_t* pOtherName  = otherDir->entryName.c_str();
        return ::wcscmp(pEntryname, pOtherName);
    }
}

bool CFDirectoryEntry::Equals(const IFObject* obj) const
{
    const IDirectoryEntry* otherDir = dynamic_cast<const IDirectoryEntry*>(obj);
    if (!otherDir) return false;
    return CompareTo(otherDir) == 0;
}

void CFDirectoryEntry::Delete()
{
    delete this;
}

int CFDirectoryEntry::GetHashCode()
{
    char* pn = WCS2MBS(entryName.c_str());
    int nhash = (int)fnv_hash((unsigned char*)pn, (int)strlen(pn));
    free(pn);
    return nhash;
}

void CFDirectoryEntry::Write(CFileStream* stream)
{
    CRWStream rw = CRWStream(stream);
    
    BYTE64 _temp;
    //  zero _temp
    ::memset(_temp.byte, 0, 64);
    unsigned short* ptrg = (unsigned short*)_temp.byte;
    const wchar_t* psrc = entryName.c_str();
    for (size_t i = 0; i <= entryName.length(); ++i)    //  Copy the ending '\0' as well
    {
        ptrg[i] = (unsigned short)psrc[i];
    }
    unsigned short len = (entryName.length() + 1) * sizeof(unsigned short);
    //unsigned short len = entryName.length() * 2 + 2;
    rw.Write(_temp.byte,64);
    
    rw.Write(len);
    
    rw.Write((unsigned char)stgType);
    rw.Write((unsigned char)stgColor);
    rw.Write(leftSibling);
    rw.Write(rightSibling);
    rw.Write(child);
    rw.Write(storageCLSID.byte,16);
    rw.Write(stateBits);
    rw.Write(creationDate.byte,8);
    rw.Write(modifyDate.byte,8);
    rw.Write(startSector);
    unsigned int nSize = (unsigned int)size;
    rw.Write(nSize);
    nSize = 0;
    rw.Write(nSize);
    
    rw.Close();
}

void CFDirectoryEntry::Read(CFileStream* stream)
{
    CRWStream rw = CRWStream(stream);
    
    BYTE64 _temp;
    rw.ReadBytes(_temp.byte, 64);
    unsigned short nameLength = rw.ReadUInt16();
    
    unsigned short* pSrc = (unsigned short*)_temp.byte;
    wchar_t buf[64];
    for (unsigned short i = 0; i < nameLength; ++i)
    {
        buf[i] = pSrc[i];
    }
    
    entryName = buf;
    
    stgType = (StgType)rw.ReadByte();
    //rw.ReadByte();stgColor = eBlack;
    stgColor = (StgColor)rw.ReadByte();
    leftSibling = rw.ReadInt32();
    rightSibling = rw.ReadInt32();
    child = rw.ReadInt32();
    
    if (stgType == eStgInvalid)
    {
        leftSibling = NOSTREAM;
        rightSibling = NOSTREAM;
        child = NOSTREAM;
    }
    
    rw.ReadBytes(storageCLSID.byte, 16);
    stateBits = rw.ReadInt32();
    rw.ReadBytes(creationDate.byte, 8);
    rw.ReadBytes(modifyDate.byte, 8);
    startSector = rw.ReadInt32();
    size = rw.ReadUInt32();//rw.ReadInt64();
    rw.ReadUInt32();
}

//static
unsigned int CFDirectoryEntry::fnv_hash(unsigned char buffer[], int length)
{
    
    unsigned int h = 2166136261;
    int i;
    
    for (i = 0; i < length; i++)
        h = (h * 16777619) ^ buffer[i];
    
    return h;
}
