/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  MemoryStream.cpp
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#include "StdAfx.h"
#include "MemoryStream.h"
#include <stdlib.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFByteArray::Init(unsigned int size) 
{
	if (size>0)
	{
		_byte = (unsigned char*)malloc(size * sizeof(unsigned char));
	}
	else
	{
		_byte = NULL;
	}
    _size = size;
    _capacity = size;
}

void CFByteArray::Write(unsigned int offset, const void *pBuf, unsigned int count)
{
    //  check if we have the capacity
    if (offset + count > _capacity)
    {
        Resize(offset + count);
    }
    ::memcpy(_byte + offset, pBuf, count);
    //  update the size
    _size = offset + count;
}

void CFByteArray::Append(const void *pBuf, unsigned int count)
{
    Write(_size, pBuf, count);
}

CFByteArray& CFByteArray::operator=(const CFByteArray &src)
{   //  clear
    Clear();
    Init(src.Length());
    //  copy the content
    ::memcpy(_byte, src.CByte(), Length());
    return *this;
}

CFByteArray::~CFByteArray()
{
    Clear();
}

void CFByteArray::Resize(unsigned int newCap)
{   //  Create a new buffer of the specified size
    void* p = malloc(newCap);
    if (p != nullptr)
    {   //  copy the original
        if (_byte != nullptr)
        {
            ::memcpy(p, _byte, std::min(newCap, _size));
            //  free the original
            ::free(_byte);
        }
        //  update the buffer
        _byte = (unsigned char*)p;
        //  don't need update _size, instead, update the capacity
        _capacity = newCap;
    }
    else
    {
        //  not enough memory!
    }
}

void CFByteArray::SetLength(unsigned int nLen)
{
    if (nLen > _capacity)
    {   //  resize
        Resize(nLen);
    }
    _size = nLen;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFBuffer::BlockCopy(const CFByteArray& src, int srcOffset, CFByteArray& dst, int dstOffset, int count)
{
	if (src.CByte() == NULL || dst.Byte() == NULL || count<=0 ||
		srcOffset<0 || dstOffset<0) return ;

	if ( (unsigned int)(srcOffset+count) > src.Length()) return;

    dst.Write(dstOffset, src.CByte() + srcOffset, count);
}

void CFBuffer::BlockCopy(const CFByteArray& src, int srcOffset, unsigned char* dst, int dstOffset, int count)
{
	if (src.CByte() == NULL || dst == NULL || count<=0 ||
		srcOffset<0 || dstOffset<0) return ;
	if ( (unsigned int)(srcOffset+count) > src.Capacity()) return;
    
	unsigned char* ps = src.CByte()+srcOffset;
	unsigned char* pd = dst+dstOffset;
	memcpy( pd, ps, (size_t)count);
}

void CFBuffer::BlockCopy(const unsigned char* src, int srcOffset, CFByteArray& dst, int dstOffset, int count)
{
	if (src == NULL || dst.Byte() == NULL || count<=0 ||
		srcOffset<0 || dstOffset<0) return ;
    
    dst.Write(dstOffset, src + srcOffset, count);
}

void CFBuffer::BlockCopy(const unsigned char* src, int srcOffset, unsigned char* dst, int dstOffset, int count)
{
	if (src == NULL || dst == NULL) return ;
	const unsigned char* ps = src+srcOffset;
	unsigned char* pd = dst+dstOffset;
	memcpy( pd, ps, (size_t)count);
}

int CFBuffer::ByteLength(const CFByteArray& _array)
{
	return _array.Length();
}

unsigned char CFBuffer::GetByte(CFByteArray& _array, int index)
{
	if (index >=0 && (unsigned int)index < _array.Length()) return *(_array.Byte()+index);
	else throw CFException("The index is out of the boundary.");
}

void CFBuffer::SetByte(CFByteArray& _array, int index, unsigned char value)
{
	if (index >=0 && (unsigned int)index < _array.Length()) *(_array.Byte()+index) = value;
	else throw CFException("The index is out of the boundary.");
}

CMemoryStream::CMemoryStream(void)
{
	 _position = 0;
}


CMemoryStream::~CMemoryStream(void)
{
}

int CMemoryStream::Read(CFByteArray& buffer, int offset, int count)
{
	int len = _MIN(count, (int)(buffer.Length()-offset));
	return Read(buffer.Byte(), offset, len);
}

int CMemoryStream::Read(unsigned char buffer[], int offset, int count)
{
	if (offset < 0 || count <=0 ||
		_position>=_buffer.Length()) return -1;

	int len = _MIN(count, (int)(_buffer.Length()-_position));

	if (len<=0) return -1;

	CFBuffer::BlockCopy(_buffer, _position, buffer, offset, len);

	_position += len;

	return len;
}

int CMemoryStream::ReadByte()
{
	if (_position<_buffer.Length()) 
	{
		unsigned char* p = _buffer.Byte()+_position;
		_position++;
		return (int)*p;
	}
	return -1;
}

int CMemoryStream::Seek(long offset, int origin)
{
	switch(origin)
	{
	case CFileStream::eBegin:
		_position = offset;
		break;
	case CFileStream::eCurrent:
		_position += offset; 
		break;
	case CFileStream::eEnd:
		_position = _buffer.Length()-offset;
		break;
	}
	return _position;
}

void CMemoryStream::SetLength(UDLONG value)
{
}

void CMemoryStream::Write(CFByteArray& buffer, int offset, int count)
{
	int len = _MIN(count, (int)(buffer.Length()-offset));
	Write(buffer.Byte(), offset, len);
}

void CMemoryStream::Write(unsigned char buffer[], int offset, int count)
{
	if (offset < 0 || count <=0 ||
		_position>=_buffer.Length()) return ;

	int len = _MIN(count, (int)(_buffer.Length()-_position));

	if (len<=0) return ;

	CFBuffer::BlockCopy(buffer, offset, _buffer, _position, len);

	_position += len;

	return ;
}

void CMemoryStream::WriteByte(unsigned char value)
{
	if (_position<_buffer.Length()) 
	{
		unsigned char* p = _buffer.Byte()+_position;
		*p = value;
		_position++;
		return ;
	}
}

void CMemoryStream::WriteTo(CFileStream* stream)
{
	if (!stream || _position>= _buffer.Length()) return ;

	int len = _buffer.Length()-_position;

	stream->Write(_buffer.Byte(), _position, len);
}

