/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

#include "StdAfx.h"

#include "RBT/RBTEnumerator.h"
#include "RBT/RBTree.h"
		
RBTEnumerator::RBTEnumerator() 
{
}

RBTEnumerator::RBTEnumerator(RBTNode* tnode, bool keys, bool ascending) 
{
	_keys       = keys;
	_ascending  = ascending;
			
    // use depth-first traversal to push nodes into stack
    // the lowest node will be at the top of the stack
    if(ascending)
	{   // find the lowest node
		while(tnode != RBTree::sentinelNode)
		{
			_stack.push(tnode);
			tnode = tnode->LeftNode();
		}
	}
	else
	{
        // the highest node will be at top of stack
		while(tnode != RBTree::sentinelNode)
		{
			_stack.push(tnode);
			tnode = tnode->RightNode();
		}
	}
			
}

IFObject* RBTEnumerator::NextElement()
{
	if(_stack.size() == 0)
	{
		throw( RBTException("Element not found"));
	}
			
	// the top of stack will always have the next item
	// get top of stack but don't remove it as the next nodes in sequence
	// may be pushed onto the top
	// the stack will be popped after all the nodes have been returned
	RBTNode* node = _stack.top();//next node in sequence
			
    if(_ascending)
    {
        if(node->RightNode() == RBTree::sentinelNode)
        {	
            // yes, top node is lowest node in subtree - pop node off stack 
            RBTNode* tn = node;
			_stack.pop();
            // peek at right node's parent 
            // get rid of it if it has already been used
            while(HasMoreElements() && _stack.top()->RightNode() == tn)
			{
                tn = _stack.top();
				_stack.pop();
			}
        }
        else
        {
            // find the next items in the sequence
            // traverse to left; find lowest and push onto stack
            RBTNode* tn = node->RightNode();
            while(tn != RBTree::sentinelNode)
            {
                _stack.push(tn);
                tn = tn->LeftNode();
            }
        }
    }
    else            // descending, same comments as above apply
    {
        if(node->LeftNode() == RBTree::sentinelNode)
        {
            // walk the tree
            RBTNode* tn = node;
			_stack.pop();
            while(HasMoreElements() && _stack.top()->LeftNode() == tn)
			{
                tn = _stack.top();
				_stack.pop();
			}
        }
        else
        {
            // determine next node in sequence
            // traverse to left subtree and find greatest node - push onto stack
            RBTNode* tn = node->LeftNode();
            while(tn != RBTree::sentinelNode)
            {
                _stack.push(tn);
                tn = tn->RightNode();
            }
        }
    }
			
	// the following is for .NET compatibility (see MoveNext())
	SetKey(node->GetKey());
	SetValue(node->GetData());

    // ******** testing only ********
    try
    {
		_parentKey = node->ParentNode()->GetKey();            // testing only
    }
    //catch(std::exception& e)
	catch(...)
    {
        _parentKey = 0;
    }
	if(node->GetColor() == 0)                     // testing only
        _Color = "Red";
    else
        _Color = "Black";
    // ******** testing only ********

    if(_keys == true) return (IFObject*)node->GetKey();
	else return (IFObject*)node->GetData();			
}
bool RBTEnumerator::MoveNext()
{
	if(HasMoreElements())
	{
		NextElement();
		return true;
	}
	return false;
}

bool RBTEnumerator::Equals(const IFObject * obj) const
{
	if (!obj) return false;
	const RBTEnumerator* pEnumer = dynamic_cast<const RBTEnumerator*> (obj);
	if (!pEnumer) return false;

	return true;
}
