/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

#include "StdAfx.h"
#include "RBT/RBTree.h"
#include "RBT/RBTNode.h"


RBTNode* RBTree::sentinelNode = new RBTNode();

RBTree::RBTree() 
{
    strIdentifier       = rand.NextToString();
    intHashCode         = rand.Next();

    // set up the sentinel node. the sentinel node is the key to a successfull
    // implementation and for understanding the red-black tree properties.
    sentinelNode->LeftNode(NULL);
    sentinelNode->RightNode(NULL);
    sentinelNode->ParentNode(NULL);
    sentinelNode->SetColor(RBTNode::BLACK);
    rbTree              = sentinelNode;
    lastNodeFound       = sentinelNode;
}
		
RBTree::RBTree(std::string strId) 
{
	intHashCode    = rand.Next();
	strIdentifier  = strId;
}
		
void RBTree::Add(IFComparable* key, IFObject* data)
{
	if(key == NULL || data == NULL)
		throw(RBTException("RBTNode key and data must not be NULL"));
			
	// traverse tree - find where node belongs
	int result			=	0;
	// create new node
	RBTNode* node	=	new RBTNode();
	RBTNode* temp	=	rbTree;				// grab the rbTree node of the tree

	while(temp != sentinelNode)
	{	// find Parent
		node->ParentNode(temp);
		result	=  key->CompareTo(temp->GetKey());
		if(result == 0)
			throw(new RBTException("A Node with the same key already exists"));
		if(result > 0)
			temp = temp->RightNode();
		else
			temp = temp->LeftNode();
	}
			
    // setup node
	node->SetKey(key);
	node->SetData(data);
    node->LeftNode(sentinelNode);
	node->RightNode(sentinelNode);

	// insert node into tree starting at parent's location
	if(node->ParentNode() != NULL)	
	{
		result	=  node->GetKey()->CompareTo(node->ParentNode()->GetKey());
		if(result > 0)
			node->ParentNode()->RightNode(node);
		else
			node->ParentNode()->LeftNode(node);
	}
	else
		rbTree = node;					// first node added

    RestoreAfterInsert(node);           // restore red-black properities

	lastNodeFound = node;
			
	intCount = intCount + 1;
}

void RBTree::RestoreAfterInsert(RBTNode* x)
{   
    // x and y are used as variable names for brevity, in a more formal
    // implementation, you should probably change the names

	RBTNode* y;

	// maintain red-black tree properties after adding x
	while(x != rbTree && x->ParentNode()->GetColor() == RBTNode::RED)
	{
		// Parent node is .Colored red; 
		if(x->ParentNode() == x->ParentNode()->ParentNode()->LeftNode())	// determine traversal path			
		{										// is it on the Left or Right subtree?
			y = x->ParentNode()->ParentNode()->RightNode();			// get uncle
			if(y!= NULL && y->GetColor() == RBTNode::RED)
			{	// uncle is red; change x's Parent and uncle to black
				x->ParentNode()->SetColor(RBTNode::BLACK);
				y->SetColor(RBTNode::BLACK);
				// grandparent must be red. Why? Every red node that is not 
				// a leaf has only black children 
				x->ParentNode()->ParentNode()->SetColor(RBTNode::RED);	
				x = x->ParentNode()->ParentNode();	// continue loop with grandparent
			}	
			else
			{
				// uncle is black; determine if x is greater than Parent
				if(x == x->ParentNode()->RightNode()) 
				{	// yes, x is greater than Parent; rotate Left
					// make x a Left child
					x = x->ParentNode();
					RotateLeft(x);
				}
				// no, x is less than Parent
				x->ParentNode()->SetColor(RBTNode::BLACK);	// make Parent black
				x->ParentNode()->ParentNode()->SetColor(RBTNode::RED);		// make grandparent black
				RotateRight(x->ParentNode()->ParentNode());					// rotate right
			}
		}
		else
		{	// x's Parent is on the Right subtree
			// this code is the same as above with "Left" and "Right" swapped
			y = x->ParentNode()->ParentNode()->LeftNode();
			if(y!= NULL && y->GetColor() == RBTNode::RED)
			{
				x->ParentNode()->SetColor(RBTNode::BLACK);
				y->SetColor(RBTNode::BLACK);
				x->ParentNode()->ParentNode()->SetColor(RBTNode::RED);
				x = x->ParentNode()->ParentNode();
			}
			else
			{
				if(x == x->ParentNode()->LeftNode())
				{
					x = x->ParentNode();
					RotateRight(x);
				}
				x->ParentNode()->SetColor(RBTNode::BLACK);
				x->ParentNode()->ParentNode()->SetColor(RBTNode::RED);
				RotateLeft(x->ParentNode()->ParentNode());
			}
		}																													
	}
	rbTree->SetColor(RBTNode::BLACK);		// rbTree should always be black
}
		
void RBTree::RotateLeft(RBTNode* x)
{
	// pushing node x down and to the Left to balance the tree. x's Right child (y)
	// replaces x (since y > x), and y's Left child becomes x's Right child 
	// (since it's < y but > x).
            
	RBTNode* y = x->RightNode();			// get x's Right node, this becomes y

	// set x's Right link
	x->RightNode(y->LeftNode());					// y's Left child's becomes x's Right child

	// modify parents
	if(y->LeftNode() != sentinelNode) 
		y->LeftNode()->ParentNode(x);				// sets y's Left Parent to x

    if(y != sentinelNode)
		y->ParentNode(x->ParentNode());			// set y's Parent to x's Parent

	if(x->ParentNode() != NULL)		
	{	// determine which side of it's Parent x was on
		if(x == x->ParentNode()->LeftNode())			
			x->ParentNode()->LeftNode(y);			// set Left Parent to y
		else
			x->ParentNode()->RightNode(y);			// set Right Parent to y
	} 
	else 
		rbTree = y;						// at rbTree, set it to y

	// link x and y 
	y->LeftNode(x);							// put x on y's Left 
	if(x != sentinelNode)						// set y as x's Parent
		x->ParentNode(y);		
}

void RBTree::RotateRight(RBTNode* x)
{
	// pushing node x down and to the Right to balance the tree. x's Left child (y)
	// replaces x (since x < y), and y's Right child becomes x's Left child 
	// (since it's < x but > y).
            
	RBTNode* y = x->LeftNode();			// get x's Left node, this becomes y

	// set x's Right link
	x->LeftNode(y->RightNode());					// y's Right child becomes x's Left child

	// modify parents
	if(y->RightNode() != sentinelNode) 
		y->RightNode()->ParentNode(x);				// sets y's Right Parent to x

    if(y != sentinelNode)
        y->ParentNode(x->ParentNode());			// set y's Parent to x's Parent

	if(x->ParentNode() != NULL)				// NULL=rbTree, could also have used rbTree
	{	// determine which side of it's Parent x was on
		if(x == x->ParentNode()->RightNode())			
			x->ParentNode()->RightNode(y);			// set Right Parent to y
		else
			x->ParentNode()->LeftNode(y);			// set Left Parent to y
	} 
	else 
		rbTree = y;						// at rbTree, set it to y

	// link x and y 
	y->RightNode(x);						// put x on y's Right
	if(x != sentinelNode)				// set y as x's Parent
		x->ParentNode(y);		
}		

IFObject* RBTree::GetData(const IFComparable* key)
{
	int result;
			
    RBTNode* treeNode = rbTree;     // begin at root
            
    // traverse tree until node is found
    while(treeNode != sentinelNode)
	{
		result = key->CompareTo(treeNode->GetKey());
		if(result == 0)
		{
			lastNodeFound = treeNode;
			return treeNode->GetData();
		}
		if(result < 0)
			treeNode = treeNode->LeftNode();
		else
			treeNode = treeNode->RightNode();
	}
			
	throw(new RBTException("RBTNode key was not found"));
}

IFComparable* RBTree::GetMinKey()
{
	RBTNode* treeNode = rbTree;
			
    if(treeNode == NULL || treeNode == sentinelNode)
		throw(new RBTException("RBT tree is empty"));
			
    // traverse to the extreme left to find the smallest key
	while(treeNode->LeftNode() != sentinelNode)
		treeNode = treeNode->LeftNode();
			
	lastNodeFound = treeNode;
			
	return treeNode->GetKey();
}

IFComparable* RBTree::GetMaxKey()
{
	RBTNode* treeNode = rbTree;
			
    if(treeNode == NULL || treeNode == sentinelNode)
        throw(new RBTException("RBT tree is empty"));

    // traverse to the extreme right to find the largest key
	while(treeNode->RightNode() != sentinelNode)
		treeNode = treeNode->RightNode();

	lastNodeFound = treeNode;

	return treeNode->GetKey();
			
}

IFObject* RBTree::GetMinValue()
{
	return GetData(GetMinKey());
}

IFObject* RBTree::GetMaxValue()
{
	return GetData(GetMaxKey());
}

RBTEnumerator* RBTree::GetEnumerator()
{
    // elements is simply a generic name to refer to the 
    // data objects the nodes contain
	return Elements(true);      
}

RBTEnumerator* RBTree::Keys()
{
	return Keys(true);
}
RBTEnumerator* RBTree::Keys(bool ascending)
{
	return new RBTEnumerator(rbTree, true, ascending);
}

RBTEnumerator* RBTree::Values()
{
	return Elements(true);
}

RBTEnumerator* RBTree::Elements()
{
	return Elements(true);
}
RBTEnumerator* RBTree::Elements(bool ascending)
{
	return new RBTEnumerator(rbTree, false, ascending);
}

bool RBTree::IsEmpty() const
{
	return (rbTree == NULL);
}

void RBTree::Remove(IFComparable* key)
{
    if(key == NULL)
        throw(new RBTException("RBTNode key is NULL"));
		
	// find node
	int	result;
	RBTNode* node;

	// see if node to be deleted was the last one found
	result = key->CompareTo(lastNodeFound->GetKey());
	if(result == 0)
		node = lastNodeFound;
	else
	{	// not found, must search		
		node = rbTree;
		while(node != sentinelNode)
		{
			result = key->CompareTo(node->GetKey());
			if(result == 0)
				break;
			if(result < 0)
				node = node->LeftNode();
			else
				node = node->RightNode();
		}

		if(node == sentinelNode)
			return;				// key not found
	}

	Delete(node);
			
	intCount = intCount - 1;
}

void RBTree::Delete(RBTNode* z)
{
	// A node to be deleted will be: 
	//		1. a leaf with no children
	//		2. have one child
	//		3. have two children
	// If the deleted node is red, the red black properties still hold.
	// If the deleted node is black, the tree needs rebalancing

	RBTNode* x = new RBTNode();	// work node to contain the replacement node
	RBTNode* y;					// work node 

	// find the replacement node (the successor to x) - the node one with 
	// at *most* one child. 
	if(z->LeftNode() == sentinelNode || z->RightNode() == sentinelNode) 
		y = z;						// node has sentinel as a child
	else 
	{
		// z has two children, find replacement node which will 
		// be the leftmost node greater than z
		y = z->RightNode();				        // traverse right subtree	
		while(y->LeftNode() != sentinelNode)		// to find next node in sequence
			y = y->LeftNode();
	}

	// at this point, y contains the replacement node. it's content will be copied 
	// to the valules in the node to be deleted

	// x (y's only child) is the node that will be linked to y's old ParentNode()-> 
    if(y->LeftNode() != sentinelNode)
        x = y->LeftNode();					
    else
        x = y->RightNode();					

	// replace x's parent with y's parent and
	// link x to proper subtree in parent
	// this removes y from the chain
	x->ParentNode(y->ParentNode());
	if(y->ParentNode() != NULL)
		if(y == y->ParentNode()->LeftNode())
			y->ParentNode()->LeftNode(x);
		else
			y->ParentNode()->RightNode(x);
	else
		rbTree = x;			// make x the root node

	// copy the values from y (the replacement node) to the node being deleted.
	// note: this effectively deletes the node. 
	if(y != z) 
	{
		z->SetKey(y->GetKey());
		z->SetData(y->GetData());
	}

	if(y->GetColor() == RBTNode::BLACK)
		RestoreAfterDelete(x);

	lastNodeFound = sentinelNode;
}

void RBTree::RestoreAfterDelete(RBTNode* x)
{
	// maintain Red-Black tree balance after deleting node 			

	RBTNode* y;

	while(x != rbTree && x->GetColor() == RBTNode::BLACK) 
	{
		if(x == x->ParentNode()->LeftNode())			// determine sub tree from parent
		{
			y = x->ParentNode()->RightNode();			// y is x's sibling 
			if(y->GetColor() == RBTNode::RED) 
			{	// x is black, y is red - make both black and rotate
				y->SetColor(RBTNode::BLACK);
				x->ParentNode()->SetColor(RBTNode::RED);
				RotateLeft(x->ParentNode());
				y = x->ParentNode()->RightNode();
			}
			if(y->LeftNode()->GetColor() == RBTNode::BLACK && 
				y->RightNode()->GetColor() == RBTNode::BLACK) 
			{	// children are both black
				y->SetColor(RBTNode::RED);		// change parent to red
				x = x->ParentNode();					// move up the tree
			} 
			else 
			{
				if(y->RightNode()->GetColor() == RBTNode::BLACK) 
				{
					y->LeftNode()->SetColor(RBTNode::BLACK);
					y->SetColor(RBTNode::RED);
					RotateRight(y);
					y = x->ParentNode()->RightNode();
				}
				y->SetColor(x->ParentNode()->GetColor());
				x->ParentNode()->SetColor(RBTNode::BLACK);
				y->RightNode()->SetColor(RBTNode::BLACK);
				RotateLeft(x->ParentNode());
				x = rbTree;
			}
		} 
		else 
		{	// right subtree - same as code above with right and left swapped
			y = x->ParentNode()->LeftNode();
			if(y->GetColor() == RBTNode::RED) 
			{
				y->SetColor(RBTNode::BLACK);
				x->ParentNode()->SetColor(RBTNode::RED);
				RotateRight (x->ParentNode());
				y = x->ParentNode()->LeftNode();
			}
			if(y->RightNode()->GetColor() == RBTNode::BLACK && 
				y->LeftNode()->GetColor() == RBTNode::BLACK) 
			{
				y->SetColor(RBTNode::RED);
				x = x->ParentNode();
			} 
			else 
			{
				if(y->LeftNode()->GetColor() == RBTNode::BLACK) 
				{
					y->RightNode()->SetColor(RBTNode::BLACK);
					y->SetColor(RBTNode::RED);
					RotateLeft(y);
					y = x->ParentNode()->LeftNode();
				}
				y->SetColor(x->ParentNode()->GetColor());
				x->ParentNode()->SetColor(RBTNode::BLACK);
				y->LeftNode()->SetColor(RBTNode::BLACK);
				RotateRight(x->ParentNode());
				x = rbTree;
			}
		}
	}
	x->SetColor(RBTNode::BLACK);
}
		
void RBTree::RemoveMin()
{
    if(rbTree == NULL)
        throw(new RBTException("RBTNode is NULL"));
            
        Remove(GetMinKey());
}

void RBTree::RemoveMax()
{
    if(rbTree == NULL)
        throw(new RBTException("RBTNode is NULL"));
            
    Remove(GetMaxKey());
}

void RBTree::Clear()
{
	rbTree      = sentinelNode;
	intCount    = 0;
}

int RBTree::Size()
{
	// number of keys
	return intCount;
}

bool RBTree::Equals(const IFObject* obj) const
{
	if(obj == NULL)
		return false;
	
	const RBTree* pTree = dynamic_cast<const RBTree*> (obj);
	if (!pTree) return false;

	if(this == obj)
		return true;
			
	return (strIdentifier == pTree->ToString());
}

int RBTree::GetHashCode()
{
	return intHashCode;
}

std::string RBTree::ToString() const
{
	return strIdentifier;
}
